"""
The classes `Token` and `Nonterm` can be subclassed and enriched
with docstrings indicating the intended grammar, and will then be
used in the parsing as part of the abstract syntax tree that is
constructed in the process.
"""

from __future__ import annotations

from mypy_extensions import mypyc_attr


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class Symbol:
    pass


class Nonterm(Symbol):
    """
    Non-terminal symbols have sets of productions associated with them.  The
    productions induce a parse forest on an input token stream.  There is
    one special non-terminal, which is denoted via the %start directive,
    whereas all other non-terminals are denoted via the %nonterm directive.
    In addition to productions (%reduce directives associated with class
    methods), the merge() method may be called during resolution of
    ambiguous parses.  See the merge() documentation for further details.

    Following are examples of how to specify non-terminal classes and their
    associated productions:

    class E(Parsing.Nonterm):
        "%start E"
        def __init__(self):
            Parsing.Nonterm.__init__(self)
            # ...

        # Productions.
        def reduceA(self, E, plus, T):
            "%reduce E plus T [split]"
            print "%r ::= %r %r %r." % (self, E, plus, T)

        def reduceB(self, T):
            "%reduce T"

    class T(Parsing.Nonterm):
            "%nonterm" # Name implicitly same as class name.
        def reduceA(self, T, star, F):
            "%reduce T star F"

        def reduceB(self, F):
            "%reduce F [p1]"

    class F(Parsing.Nonterm):
        "%nonterm F [p2]"
        def reduceA(self, lparen, E, rparen):
            "%reduce lparen E rparen"

        def reduceB(self, id):
            "%reduce id"
    """

    def merge(self, other: Nonterm) -> Nonterm:
        """
        Merging happens when there is an ambiguity in the input that allows
        non-terminals to be part of multiple overlapping series of
        reductions.  If no merge() method is specified, the parser will
        throw a syntax error upon encountering an ambiguity that confounds
        reduction processing.  However, it may be useful to either discard
        one of the possible parses, or to explicitly record the ambiguity in
        the data structures being created during parsing.  In both of these
        cases, the non-terminal-specific merge() is the place to do the
        work; merge() returns an object that is stored by the parser onto
        the parse stack.  In the case where merge() discards one of the
        possible parses, it need only return the parse that is to be
        preserved (self or other).

        If multiple merges are necessary, they cause a series of merge()
        calls.  The first alternative (self) may be the result of a previous
        merge() call, whereas other will not have not been merged yet
        (unless as the result of merging further down in the parse forest).

        The alternative that is discarded is never touched by the parser
        again, so if any immediate cleanup is necessary, it should be done
        in merge().
        """
        raise SyntaxError(
            "No merge() for %r; merging %r <--> %r" % (type(self), self, other)
        )


class Token(Symbol):
    """
    Tokens are terminal symbols.  The parser is fed Token instances, which
    is what drives parsing.  Typically, the user will define a class that
    subclasses Parsing.Token and implement parser-specific machinery there,
    then derive all actual token types from that class.

    class Token(Parsing.Token):
        def __init__(self, parser):
            Parsing.Token.__init__(self, parser)
            # ...

    class Plus(Token):
        "%token plus [p1]"

    class star(Token):
        "%token star [p2]" # Name implicitly same as class name.

    class lparen(Token):
        "%token [split]"

    class rparen(Token):
        "%token [none]" # [none] not necessary, since it's the default.

    class id(Token):
        "%token" """


class Precedence:
    """
    Precedences can be associated with tokens, non-terminals, and
    productions.  Precedence isn't as important for GLR parsers as for LR
    parsers, since GLR parsing allows for parse-time resolution of
    ambiguity.  Still, precedence can be useful for reducing the volume of
    ambiguities that must be dealt with at run-time.

    There are five precedence types: %fail, %nonassoc, %left, %right, and
    %split.  Each precedence can have relationships with other precedences:
    <, >, or =.  These relationships specify a directed acyclic graph (DAG),
    which is used to compute the transitive closures of relationships among
    precedences.  If no path exists between two precedences that are
    compared during conflict resolution, parser generation fails.  < and >
    are reflexive; it does not matter which is used.  Conceptually, the =
    relationship causes precedences to share a node in the DAG.

    During conflict resolution, an error results if no path exists in the
    DAG between the precedences under consideration.  When such a path
    exists, the highest precedence non-terminal or production takes
    precedence.  Associativity only comes into play for shift/reduce
    conflicts, where the terminal and the production have equivalent
    precedences (= relationship).  In this case, the non-terminal's
    associativity determines how the conflict is resolved.

    The %fail and %split associativities are special because they can be
    mixed with other associativities.  During conflict resolution, if
    another action has non-%fail associativity, then the %fail (lack of)
    associativity is overridden.  Similarly, %split associativity overrides
    any other associativity.  In contrast, any mixture of associativity
    between %nonassoc/%left/%right causes an unresolvable conflict.

        %fail : Any conflict is a parser-generation-time error.

                A pre-defined precedence, [none], is provided.  It has
                %fail associativity, and has no pre-defined precedence
                relationships.

    %nonassoc : Resolve shift/reduce conflicts by removing both
                possibilities, thus making conflicts a parse-time error.

        %left : Resolve shift/reduce conflicts by reducing.

        %right : Resolve shift/reduce conflicts by shifting.

        %split : Do not resolve conflicts; the GLR algorithm will split
                the parse stack when necessary.

                A pre-defined precedence, [split], is provided.  It has
                %split associativity, and has no pre-defined precedence
                relationships.

    By default, all symbols have [none] precedence.  Each production
    inherits the precedence of its left-hand-side nonterminal's precedence
    unless a precedence is manually specified for the production.

    Following are some examples of how to specify precedence classes:

    class P1(Parsing.Precedence):
        "%split p1"

    class p2(Parsing.Precedence):
        "%left" # Name implicitly same as class name.

    class P3(Parsing.Precedence):
        "%left p3 >p2" # No whitespace is allowed between > and p2.

    class P4(Parsing.Precedence):
        "%left p4 =p3" # No whitespace is allowed between = and p3.
    """



import unittest
import parsing


class TestParsing(unittest.TestCase):
    def test_basic_a(self):
        class TestParser(parsing.Lr):
            def __init__(self, spec):
                parsing.Lr.__init__(self, spec)

        from parsing.tests.specs import a

        spec = parsing.Spec(a)

        parser = TestParser(spec)
        parser.token(a.TokenId())
        parser.token(a.TokenStar())
        parser.token(a.TokenId())
        parser.token(a.TokenPlus())
        parser.token(a.TokenId())
        parser.eoi()
        self.assertEqual(len(parser.start), 1)
        self.assertEqual(parser.start[0].val, "[[ID * ID] + ID]")

        parser = TestParser(spec)
        parser.token(a.TokenId())
        parser.token(a.TokenPlus())
        parser.token(a.TokenId())
        parser.token(a.TokenStar())
        parser.token(a.TokenId())
        parser.eoi()
        self.assertEqual(len(parser.start), 1)
        self.assertEqual(parser.start[0].val, "[ID + [ID * ID]]")

        parser = TestParser(spec)
        parser.token(a.TokenId())
        parser.token(a.TokenStar())
        parser.token(a.TokenLparen())
        parser.token(a.TokenId())
        parser.token(a.TokenPlus())
        parser.token(a.TokenId())
        parser.token(a.TokenRparen())
        parser.eoi()
        self.assertEqual(len(parser.start), 1)
        self.assertEqual(parser.start[0].val, "[ID * ([ID + ID])]")

    def test_basic_b(self):
        class TestParser(parsing.Glr):
            def __init__(self, spec):
                parsing.Glr.__init__(self, spec)

        from parsing.tests.specs import b

        spec = parsing.Spec(b, skinny=False)

        parser = TestParser(spec)
        parser.token(b.id())
        parser.token(b.star())
        parser.token(b.id())
        parser.token(b.plus())
        parser.token(b.id())
        parser.eoi()
        self.assertEqual(len(parser.start), 1)
        self.assertEqual(parser.start[0].val, "[[ID * ID] + ID]")

        parser = TestParser(spec)
        parser.token(b.id())
        parser.token(b.plus())
        parser.token(b.id())
        parser.token(b.star())
        parser.token(b.id())
        parser.eoi()
        self.assertEqual(len(parser.start), 1)
        self.assertEqual(parser.start[0].val, "[ID + [ID * ID]]")

        parser = TestParser(spec)
        parser.token(b.id())
        parser.token(b.star())
        parser.token(b.lparen())
        parser.token(b.id())
        parser.token(b.plus())
        parser.token(b.id())
        parser.token(b.rparen())
        parser.eoi()
        self.assertEqual(len(parser.start), 1)
        self.assertEqual(parser.start[0].val, "[ID * ([ID + ID])]")

    def test_basic_d(self):
        class TestParser(parsing.Glr):
            def __init__(self, spec):
                parsing.Glr.__init__(self, spec)

        from parsing.tests.specs import d

        spec = parsing.Spec(d, skinny=False)

        parser = TestParser(spec)
        parser.token(d.id())
        parser.token(d.star())
        parser.token(d.id())
        parser.token(d.plus())
        parser.token(d.id())
        parser.token(d.star())
        parser.token(d.id())
        parser.eoi()

        self.assertEqual(len(parser.start), 1)
        self.assertEqual(parser.start[0].val, "[[ID * ID] + [ID * ID]]")

    def test_basic_h(self):
        class TestGlrParser(parsing.Glr):
            def __init__(self, spec):
                parsing.Glr.__init__(self, spec)

        from parsing.tests.specs import h

        spec = parsing.Spec(h, skinny=False)

        parser = TestGlrParser(spec)
        parser.token(h.TokenI())
        parser.token(h.TokenPlus())
        parser.token(h.TokenI())
        parser.token(h.TokenStar())
        parser.token(h.TokenI())
        parser.eoi()
        self.assertEqual(len(parser.start), 1)
        self.assertEqual(repr(parser.start[0]), "(i + (i * i))")

    def test_basic_i(self):
        class TestGlrParser(parsing.Glr):
            def __init__(self, spec):
                parsing.Glr.__init__(self, spec)

        from parsing.tests.specs import i

        self.assertRaises(parsing.SpecError, parsing.Spec, i)

    def test_basic_pickle(self):
        class TestGlrParser(parsing.Glr):
            def __init__(self, spec):
                parsing.Glr.__init__(self, spec)

        from parsing.tests.specs import b

        spec = parsing.Spec(b, skinny=False)
        import pickle

        specPickle = pickle.dumps(spec)
        spec2 = pickle.loads(specPickle)

        parser = TestGlrParser(spec2)
        parser.token(b.id())
        parser.token(b.star())
        parser.token(b.id())
        parser.token(b.plus())
        parser.token(b.id())
        parser.eoi()
        self.assertEqual(len(parser.start), 1)
        self.assertEqual(parser.start[0].val, "[[ID * ID] + ID]")


if __name__ == "__main__":
    unittest.main()



import parsing


class TokenI(parsing.Token):
    "%token i"


class TokenPlus(parsing.Token):
    "%token plus [split]"


class TokenStar(parsing.Token):
    "%token star [split]"


class S(parsing.Nonterm):
    "%start"

    def __repr__(self):
        return "%r" % self.E

    def reduce(self, E):
        "%reduce E"
        self.E = E


class E(parsing.Nonterm):
    "%nonterm"

    def __repr__(self):
        if self.variant == "i":
            return "i"
        elif self.variant == "plus":
            return "(%r + %r)" % (self.EA, self.EB)
        elif self.variant == "star":
            return "(%r * %r)" % (self.EA, self.EB)
        else:
            assert False

    def merge(self, other):
        if self.variant == "plus":
            return self
        elif other.variant == "plus":
            return other
        else:
            return self

    def reduceI(self, i):
        "%reduce i"
        self.variant = "i"
        self.i = i

    def reducePlus(self, EA, plus, EB):
        "%reduce E plus E [split]"
        self.variant = "plus"
        self.EA = EA
        self.plus = plus
        self.EB = EB

    def reduceStar(self, EA, star, EB):
        "%reduce E star E [split]"
        self.variant = "star"
        self.EA = EA
        self.star = star
        self.EB = EB



import os.path
import subprocess
import sys
import unittest


def find_root():
    return os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )


class TestCodeQuality(unittest.TestCase):
    def test_flake8(self):
        rootpath = find_root()

        try:
            import flake8  # NoQA
        except ImportError:
            raise unittest.SkipTest("flake8 moudule is missing")

        for subdir in ["parsing"]:
            try:
                subprocess.check_output(
                    [sys.executable, "-m", "flake8", subdir], cwd=rootpath
                )
            except subprocess.CalledProcessError as ex:
                output = ex.output.decode()
                raise AssertionError(
                    "flake8 validation failed:\n{}".format(output)
                )

    def test_mypy(self):
        rootpath = find_root()
        config_path = os.path.join(rootpath, "pyproject.toml")
        if not os.path.exists(config_path):
            raise RuntimeError("could not locate pyproject.toml file")

        try:
            import mypy  # NoQA
        except ImportError:
            raise unittest.SkipTest("mypy module is missing")

        try:
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "mypy",
                    "--config-file",
                    config_path,
                    "parsing",
                ],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=rootpath,
            )
        except subprocess.CalledProcessError as ex:
            output = ex.stdout.decode()
            if ex.stderr:
                output += "\n\n" + ex.stderr.decode()
            raise AssertionError(
                f"mypy validation failed:\n{output}"
            ) from None


if __name__ == "__main__":
    unittest.main()



# ============================================================================
# Copyright (c) 2007 Jason Evans <jasone@canonware.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ============================================================================
"""
This module contains classes that are used in the specification of grammars.
"""

from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
)

from mypy_extensions import mypyc_attr

import re
import sys
from parsing.ast import Token, Nonterm
from parsing.errors import SpecError
from parsing import introspection

if TYPE_CHECKING:
    import types


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class PrecedenceSpec:
    assoc_tok_re: ClassVar[re.Pattern[str]] = re.compile(
        r"([<>=])([A-Za-z]\w*)"
    )

    def __init__(
        self,
        name: str,
        assoc: str,
        relationships: Mapping[str, str],
    ) -> None:
        assert assoc in ["fail", "nonassoc", "left", "right", "split"]

        self.name = name
        self.assoc = assoc
        self.relationships = relationships  # Raw relationships specification.

        # Precedences that have equivalent precedence.
        self.equiv: set[PrecedenceSpec] = set((self,))
        # Precedences that have higher precedence.
        self.dominators: set[PrecedenceSpec] = set()

    def __repr__(self) -> str:
        equiv = [prec.name for prec in self.equiv]
        equiv.sort()
        domin = [prec.name for prec in self.dominators]
        domin.sort()
        return "[%%%s %s ={%s} <{%s}]" % (
            self.assoc,
            self.name,
            ",".join(equiv),
            ",".join(domin),
        )


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class PrecedenceRef(PrecedenceSpec):
    def __init__(self, name: str) -> None:
        super().__init__(name, "fail", {})


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class SymbolSpec:
    name: str
    prec: PrecedenceSpec
    firstSet: set[SymbolSpec]
    followSet: set[SymbolSpec]

    def __init__(self, name: str, prec: PrecedenceSpec) -> None:
        self.name = name
        self.prec = prec
        self.firstSet = set()
        self.followSet = set()

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return repr(self)

    def firstSetMerge(self, sym: SymbolSpec) -> bool:
        if sym not in self.firstSet:
            self.firstSet.add(sym)
            return False
        else:
            return True

    def followSetMerge(self, set: Iterable[SymbolSpec]) -> bool:
        ret = True
        for sym in set:
            if sym != epsilon and sym not in self.followSet:
                self.followSet.add(sym)
                ret = False
        return ret


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class NontermSpec(SymbolSpec):
    token_re: ClassVar[re.Pattern[str]] = re.compile(r"([A-Za-z]\w*)")
    precedence_tok_re: ClassVar[re.Pattern[str]] = re.compile(
        r"\[([A-Za-z]\w*)\]"
    )

    def __init__(
        self,
        nontermType: Type[Nonterm],
        name: str,
        qualified: str,
        prec: PrecedenceSpec,
    ) -> None:
        super().__init__(name, prec)
        self.qualified = qualified
        self.nontermType = nontermType
        self.productions: set[Production] = set()

    @classmethod
    def from_class(
        cls,
        nt_subclass: type,
        name: Optional[str] = None,
        module: Optional[types.ModuleType] = None,
    ) -> Tuple[NontermSpec, bool]:
        if name is None:
            name = nt_subclass.__name__
        if module is None:
            module_name = nt_subclass.__module__
        else:
            module_name = module.__name__
        if nt_subclass.__doc__ is not None:
            dirtoks = introspection.parse_docstring(nt_subclass.__doc__)
        else:
            dirtoks = ("%nonterm", name)
        is_start = dirtoks[0] == r"%start"
        symbol_name = None
        prec = None
        i = 1
        while i < len(dirtoks):
            tok = dirtoks[i]
            m = NontermSpec.precedence_tok_re.match(tok)
            if m:
                if i < len(dirtoks) - 1:
                    raise SpecError(
                        "Precedence must come last in "
                        "non-terminal specification: %s" % nt_subclass.__doc__
                    )
                prec = PrecedenceRef(m.group(1))
            else:
                m = NontermSpec.token_re.match(tok)
                if m:
                    symbol_name = m.group(1)
                else:
                    raise SpecError(
                        "Invalid non-terminal specification: %s"
                        % nt_subclass.__doc__
                    )
            i += 1
        if symbol_name is None:
            symbol_name = name
        if prec is None:
            prec = PrecedenceRef("none")

        nonterm = NontermSpec(
            nt_subclass, symbol_name, f"{module_name}.{name}", prec
        )
        return nonterm, is_start


# AKA terminal symbol.
@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class TokenSpec(SymbolSpec):
    def __init__(
        self,
        tokenType: Type[Token],
        name: str,
        prec: PrecedenceSpec,
    ) -> None:
        super().__init__(name, prec)
        self.tokenType = tokenType


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class Item:
    production: Production
    dotPos: int
    symbol: Optional[SymbolSpec]

    def __init__(
        self,
        production: Production,
        dotPos: int,
    ) -> None:
        self.production = production
        self.dotPos = dotPos
        rhsLen = len(production.rhs)
        if dotPos == rhsLen:
            self.symbol = None
        elif dotPos < rhsLen:
            self.symbol = production.rhs[dotPos]
        else:
            raise AssertionError("dotPos outside of RHS!")

    def __repr__(self) -> str:
        strs = []
        strs.append("[%r ::=" % self.production.lhs)
        assert self.dotPos <= len(self.production.rhs)
        i = 0
        while i < self.dotPos:
            strs.append(" %r" % self.production.rhs[i])
            i += 1
        strs.append(" *")
        while i < len(self.production.rhs):
            strs.append(" %r" % self.production.rhs[i])
            i += 1
        strs.append(".] [%s]" % (self.production.prec.name,))

        return "".join(strs)

    def lr0__repr__(self) -> str:
        strs = []
        strs.append("%r ::=" % self.production.lhs)
        assert self.dotPos <= len(self.production.rhs)
        i = 0
        while i < self.dotPos:
            strs.append(" %r" % self.production.rhs[i])
            i += 1
        strs.append(" *")
        while i < len(self.production.rhs):
            strs.append(" %r" % self.production.rhs[i])
            i += 1
        strs.append(". [%s]" % self.production.prec.name)

        return "".join(strs)


_item_cache: dict[tuple[Production, int], Item] = {}


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class Production:
    def __init__(
        self,
        method: Callable[..., Nonterm | None],
        qualified: str,
        prec: PrecedenceSpec,
        lhs: NontermSpec,
        rhs: Tuple[SymbolSpec, ...],
    ) -> None:
        self.method = method
        self.qualified = qualified
        self.prec = prec
        self.lhs = lhs
        self.rhs = rhs

    def __getstate__(
        self,
    ) -> Tuple[str, PrecedenceSpec, NontermSpec, Tuple[SymbolSpec, ...]]:
        return (self.qualified, self.prec, self.lhs, self.rhs)

    def __setstate__(
        self,
        data: Tuple[str, PrecedenceSpec, NontermSpec, Tuple[SymbolSpec, ...]],
    ) -> None:
        # Convert qualified name to a function reference.
        (self.qualified, self.prec, self.lhs, self.rhs) = data
        elms = self.qualified.split(".")
        assert len(elms) > 1
        module = sys.modules[elms[0]]
        method = module.__dict__[elms[1]]
        for elm in elms[2:]:
            method = method.__dict__[elm]
        self.method = method

    def __repr__(self) -> str:
        return "%r ::= %s. [%s]" % (
            self.lhs,
            " ".join(["%r" % elm for elm in self.rhs]),
            self.prec.name,
        )

    def item(self, dotPos: int) -> Item:
        key = (self, dotPos)
        item = _item_cache.get(key)
        if item is None:
            item = Item(self, dotPos)
            _item_cache[key] = item
        return item

    # Optional callback method.
    #
    # Called when a production is reduced.
    def reduce(self, lhs: NontermSpec, *rhs: SymbolSpec) -> None:
        pass


# <$>.
class EndOfInput(Token):
    pass


class EndOfInputSpec(TokenSpec):
    def __init__(self) -> None:
        super().__init__(EndOfInput, "<$>", PrecedenceRef("none"))


eoi = EndOfInputSpec()


# <e>.
class Epsilon(Token):
    pass


class EpsilonSpec(TokenSpec):
    def __init__(self) -> None:
        super().__init__(Epsilon, "<e>", PrecedenceRef("none"))


epsilon = EpsilonSpec()


class NontermStart(Nonterm):
    def reduce(self, userStartSym: SymbolSpec, eoi: EndOfInputSpec) -> None:
        pass


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class Action:
    """
    Abstract base class, subclassed by {Shift,Reduce}Action."""

    def __init__(self) -> None:
        pass


class ShiftAction(Action):
    """
    Shift action, with assocated nextState."""

    def __init__(self, nextState: int) -> None:
        super().__init__()
        self.nextState = nextState

    def __repr__(self) -> str:
        return "[shift %r]" % self.nextState

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ShiftAction):
            return False
        if self.nextState != other.nextState:
            return False
        return True


class ReduceAction(Action):
    """
    Reduce action, with associated production."""

    def __init__(self, production: Production) -> None:
        super().__init__()
        self.production = production

    def __repr__(self) -> str:
        return "[reduce %r]" % self.production

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ReduceAction):
            return False
        if self.production != other.production:
            return False
        return True


ActionState = Dict[SymbolSpec, List[Action]]
GotoState = Dict[SymbolSpec, int]



import parsing


# Precedences.
class p1(parsing.Precedence):
    "%left"


class p2(parsing.Precedence):
    "%left >p1"


# Tokens.
class plus(parsing.Token):
    "%token [p1]"


class star(parsing.Token):
    "%token [p2]"


class id(parsing.Token):
    "%token"


# Non-terminal definitions.
class S(parsing.Nonterm):
    "%start"

    def reduce(self, E):
        "%reduce E"
        self.val = E.val


class E(parsing.Nonterm):
    "%nonterm"

    def reduceId(self, id):
        "%reduce id"
        self.val = "ID"

    def reducePlus(self, EA, plus, EB):
        "%reduce E plus E [p1]"
        self.val = "[%s + %s]" % (EA.val, EB.val)

    def reduceStar(self, EA, star, EB):
        "%reduce E star E [p2]"
        self.val = "[%s * %s]" % (EA.val, EB.val)



from __future__ import annotations
from typing import Iterator, List

from parsing.ast import Symbol, Nonterm, Token
from parsing.errors import UnexpectedToken
from parsing.grammar import (
    Production,
    SymbolSpec,
    TokenSpec,
    EndOfInput,
    ShiftAction,
    ReduceAction,
)
from parsing.lrparser import Lr


class GssPathStep:
    pass


class Gssn(GssPathStep):
    """Graph-structured stack node."""

    def __init__(
        self, below: Gssn | None, value: Symbol | None, nextState: int
    ) -> None:
        self._edges: list[Gsse] = []
        if below is not None and value is not None:
            Gsse(below, self, value)
        self.nextState = nextState

    def __repr__(self) -> str:
        return "[%d]" % self.nextState

    @property
    def edge(self) -> Gsse:
        assert len(self._edges) == 1
        return self._edges[0]

    def edges(self) -> Iterator[Gsse]:
        for edge in self._edges:
            yield edge

    def nodes(self) -> Iterator[Gssn]:
        for edge in self._edges:
            yield edge.node

    # Iterate over all paths of length pathLen.  Path length is measured as the
    # number of edges in the path, so a path of length 0 still consists of a
    # single node.
    #
    # Each path is encoded as a list that alternates between nodes and edges,
    # where the first and last elements are always nodes.
    #
    # <e>-grammars can cause cycles, which requires that we avoid infinite
    # recursion.
    def paths(
        self, pathLen: int | None = None
    ) -> Iterator[tuple[GssPathStep, ...]]:
        assert pathLen is None or isinstance(pathLen, int) and pathLen >= 0

        for path in self._pathsRecurse(pathLen, []):
            yield path

    def _pathsRecurse(
        self, pathLen: int | None, path: list[GssPathStep]
    ) -> Iterator[tuple[GssPathStep, ...]]:
        path.insert(0, self)
        if pathLen is None and len(self._edges) == 0:
            yield tuple(path[:])
        elif pathLen is not None and len(path) - 1 == pathLen * 2:
            yield tuple(path[:])
        else:
            for edge in self.edges():
                # Avoid infinite recursion due to <e>-production cycles.
                if len(path) < 3 or edge != path[1]:
                    path.insert(0, edge)
                    for x in edge.node._pathsRecurse(pathLen, path):
                        yield x
                    path.pop(0)
        path.pop(0)


class Gsse(GssPathStep):
    """Graph-structured stack edge."""

    def __init__(self, below: Gssn, above: Gssn, value: Symbol) -> None:
        self.node = below
        above._edges.append(self)
        self.value = value

    def __repr__(self) -> str:
        return "{%r}" % self.value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Gsse):
            return NotImplemented
        else:
            return self.node == other.node and self.value == other.value


#
# End graph-structured stack (GSS) classes.
# ========================================================================


class Glr(Lr):
    """
    GLR parser.  The Glr class uses a Spec instance in order to parse input
    that is fed to it via the token() method, and terminated via the eoi()
    method."""

    def reset(self) -> None:
        self._start = None

        # Initialize with a stack that is in the start state.
        self._gss: List[Gssn] = []
        top = Gssn(None, None, 0)
        self._gss.append(top)

    def token(self, token: Token) -> None:
        """
        Feed a token to the parser."""
        if self.verbose:
            print("%s" % ("-" * 80))
            print("INPUT: %r" % token)
        tokenSpec = self._spec.sym_spec(token)
        self._act(token, tokenSpec)  # type: ignore
        if len(self._gss) == 0:
            raise UnexpectedToken("Unexpected token: %r" % token)

    def eoi(self) -> None:
        """
        Signal end-of-input to the parser."""
        token = EndOfInput()
        self.token(token)

        # Gather the start symbols from the stacks.
        self._start = []
        for top in self._gss:
            for path in top.paths():
                assert len(path) == 5
                if self.verbose:
                    print("   --> accept %r" % path)
                edge = path[1]
                assert isinstance(edge, Gsse)
                assert isinstance(edge.value, Nonterm)
                assert (
                    self._spec.sym_spec(edge.value) == self._start_sym
                )
                self._start.append(edge.value)

        if len(self._start) == 0:
            raise UnexpectedToken("Unexpected end of input")

        if self.verbose:
            print("Start: %r" % self._start)
            print("%s" % ("-" * 80))

    def _act(self, sym: Token, symSpec: TokenSpec) -> None:
        self._reductions(sym, symSpec)
        self._shifts(sym, symSpec)

    def _reductions(self, sym: Token, symSpec: TokenSpec) -> None:
        # epsilons is a dictionary that maps production-->[tops].  The purpose
        # is to avoid repeating the same epsilon production on a particular
        # stack top.  Ordinary productions do not require this care because we
        # can notice when a path has already been used for a production.
        epsilons = {}

        if self.verbose:
            nReduces = 0

        # Enqueue work.
        workQ: list[tuple[tuple[GssPathStep, ...], Production]] = []
        i = 0
        while i < len(self._gss):
            top = self._gss[i]
            if symSpec not in self._action[top.nextState]:
                # Unexpected token for this stack.
                self._gss.pop(i)
            else:
                for action in self._action[top.nextState][symSpec]:
                    if type(action) == ReduceAction:
                        if len(action.production.rhs) == 0:
                            if action.production not in epsilons:
                                assert (
                                    len([path for path in top.paths(0)]) == 1
                                )
                                path = [p for p in top.paths(0)][0]
                                epsilons[action.production] = [top]
                                workQ.append((path, action.production))
                                if self.verbose:
                                    print(
                                        "   --> enqueue(a) %r"
                                        % action.production
                                    )
                                    print("                  %r" % path)
                            elif top not in epsilons[action.production]:
                                assert (
                                    len([path for path in top.paths(0)]) == 1
                                )
                                path = [p for p in top.paths(0)][0]
                                epsilons[action.production].append(top)
                                workQ.append((path, action.production))
                                if self.verbose:
                                    print(
                                        "   --> enqueue(b) %r"
                                        % action.production
                                    )
                                    print("                  %r" % path)
                        else:
                            # Iterate over all reduction paths through stack
                            # and enqueue them.
                            for path in top.paths(len(action.production.rhs)):
                                workQ.append((path, action.production))
                                if self.verbose:
                                    print(
                                        "   --> enqueue(c) %r"
                                        % action.production
                                    )
                                    print("                  %r" % path)
                i += 1

        # Process the work queue.
        while len(workQ) > 0:
            (path, production) = workQ.pop(0)

            if self.verbose:
                print("   --> reduce %r" % production)
                print("              %r" % path)
                nReduces += 1

            self._glr_reduce(workQ, epsilons, path, production, symSpec)

        if self.verbose:
            if nReduces > 0:
                self._printStack()

    def _glr_reduce(
        self,
        workQ: list[tuple[tuple[GssPathStep, ...], Production]],
        epsilons: dict[Production, list[Gssn]],
        path: tuple[GssPathStep, ...],
        production: Production,
        symSpec: SymbolSpec,
    ) -> None:
        assert len(path[1::2]) == len(production.rhs)

        # Build the list of RHS semantic values to pass to the reduction
        # action.
        rhs = [edge.value for edge in path[1::2]]  # type: ignore

        # Call the user reduction method.
        r = self._production(production, rhs)

        below = path[0]
        assert isinstance(below, Gssn)
        done = False
        for top in self._gss:
            if top.nextState == self._goto[below.nextState][production.lhs]:
                # top is compatible with the reduction result we want to add to
                # the set of stack tops.
                for edge in top.edges():
                    if edge.node == below:
                        nonterm = edge.value
                        assert isinstance(nonterm, Nonterm)
                        # There is already a below<--top link, so merge
                        # competing interpretations.
                        if self.verbose:
                            print("   --> merge %r <--> %r" % (nonterm, r))
                        value = production.lhs.nontermType.merge(nonterm, r)
                        if self.verbose:
                            if value == edge.value:
                                print(
                                    "             %s"
                                    % ("-" * len("%r" % nonterm))
                                )
                            else:
                                print(
                                    "             %s      %s"
                                    % (
                                        (" " * len("%r" % nonterm)),
                                        "-" * len("%r" % r),
                                    )
                                )
                        edge.value = value
                        done = True
                        break
                if not done:
                    # Create a new below<--top link.
                    edge = Gsse(below, top, r)
                    if self.verbose:
                        print("   --> shift(b) %r" % top)

                    # Enqueue reduction paths that were created as a result of
                    # the new link.
                    self._enqueueLimitedReductions(
                        workQ, epsilons, edge, symSpec
                    )
                    done = True
                break
        if not done:
            # There is no compatible stack top, so create a new one.
            top = Gssn(below, r, self._goto[below.nextState][production.lhs])
            self._gss.append(top)
            if self.verbose:
                print(
                    "   --> shift(c) %r"
                    % self._goto[below.nextState][production.lhs]
                )
            self._enqueueLimitedReductions(workQ, epsilons, top.edge, symSpec)

    # Enqueue paths that incorporate edge.
    def _enqueueLimitedReductions(
        self,
        workQ: list[tuple[tuple[GssPathStep, ...], Production]],
        epsilons: dict[Production, list[Gssn]],
        edge: Gsse,
        symSpec: SymbolSpec,
    ) -> None:
        gotos = self._goto

        for top in self._gss:
            if symSpec in self._action[top.nextState]:
                for action in self._action[top.nextState][symSpec]:
                    if type(action) == ReduceAction:
                        if len(action.production.rhs) == 0:
                            if (
                                gotos[top.nextState][action.production.lhs]
                                == top.nextState
                            ):
                                # Do nothing, since enqueueing a reduction
                                # would result in performing the same reduction
                                # twice.
                                pass
                            elif action.production not in epsilons:
                                p = (top,)
                                epsilons[action.production] = [top]
                                workQ.append((p, action.production))
                                if self.verbose:
                                    print(
                                        "   --> enqueue(d) %r"
                                        % action.production
                                    )
                                    print("                  %r" % p)
                            elif top not in epsilons[action.production]:
                                path = (top,)
                                epsilons[action.production].append(top)
                                workQ.append((path, action.production))
                                if self.verbose:
                                    print(
                                        "   --> enqueue(e) %r"
                                        % action.production
                                    )
                                    print("                  %r" % path)
                        else:
                            # Iterate over all reduction paths through stack
                            # and enqueue them if they incorporate edge.
                            for rp in top.paths(len(action.production.rhs)):
                                if edge in rp[1::2]:
                                    workQ.append((rp, action.production))
                                    if self.verbose:
                                        print(
                                            "   --> enqueue(f) %r"
                                            % action.production
                                        )
                                        print("                  %r" % rp)

    def _shifts(self, sym: Token, symSpec: TokenSpec) -> None:
        prevGss = self._gss
        self._gss = []

        if self.verbose:
            nShifts = 0

        for topA in prevGss:
            if symSpec in self._action[topA.nextState]:
                for action in self._action[topA.nextState][symSpec]:
                    if type(action) == ShiftAction:
                        merged = False
                        for topB in self._gss:
                            if topB.nextState == topA.nextState:
                                Gsse(topA, topB, sym)
                                merged = True
                                break
                        if not merged:
                            top = Gssn(topA, sym, action.nextState)
                            self._gss.append(top)
                            if self.verbose:
                                print("   --> shift(a) %d" % action.nextState)
                                nShifts += 1
        if self.verbose:
            if nShifts > 0:
                self._printStack()

    def _printStack(self) -> None:
        i = 0
        for top in self._gss:
            for path in top.paths():
                if i == 0:
                    print("STK 0:", end=" ")
                else:
                    print("    %d:" % i, end=" ")
                for elm in path:
                    print("%r" % elm, end=" ")
                print()
                i += 1



import parsing


# Tokens.
class plus(parsing.Token):
    "%token"


class star(parsing.Token):
    "%token"


class lparen(parsing.Token):
    "%token"


class rparen(parsing.Token):
    "%token"


class id(parsing.Token):
    "%token"


# Non-terminal definitions.
class E(parsing.Nonterm):
    "%start"

    def reduceA(self, E, plus, T):
        "%reduce E plus T"
        self.val = "[%s + %s]" % (E.val, T.val)

    def reduceB(self, T):
        "%reduce T"
        self.val = T.val


class T(parsing.Nonterm):
    "%nonterm"

    def reduceA(self, T, star, F):
        "%reduce T star F"
        self.val = "[%s * %s]" % (T.val, F.val)

    def reduceB(self, F):
        "%reduce F"
        self.val = F.val


class F(parsing.Nonterm):
    "%nonterm"

    def reduceA(self, lparen, E, rparen):
        "%reduce lparen E rparen"
        self.val = "(%s)" % (E.val,)

    def reduceB(self, id):
        "%reduce id"
        self.val = "ID"



"""
The Parsing module implements the following exception classes:

  * Exception
  * SpecError
  * UnexpectedToken
"""


# ============================================================================
# Begin exceptions.
#
class AnyException(Exception):
    """
    Top-level class for all exceptions thrown within the Parsing module.
    Needed for compatibility with old code - do not use.
    """


class ParsingError(AnyException):
    """
    Top level Parsing exception class, from which we derive all exceptions that
    occur during the parsing of an input string.
    """


class SpecError(AnyException):
    """
    Specification error exception.  SpecError arises when the Spec
    introspection machinery detects an error either during docstring parsing
    or parser specification generation.
    """


class UnexpectedToken(ParsingError):
    """
    Parser syntax error.  UnexpectedToken arises when a Parser instance detects
    a syntax error according to the Spec it is using, for the input being
    fed to it.
    """


#
# End exceptions.
# ============================================================================



from __future__ import annotations

from parsing.ast import Symbol, Nonterm, Token
from parsing.errors import UnexpectedToken
from parsing.grammar import (
    Production,
    TokenSpec,
    EndOfInput,
    Epsilon,
    ShiftAction,
    ReduceAction,
)
from parsing.interfaces import Parser, Spec


class Lr(Parser):
    """
    LR(1) parser.  The Lr class uses a Spec instance in order to parse
    input that is fed to it via the token() method, and terminated via the
    eoi() method.
    """

    _spec: Spec
    _start: list[Symbol] | None
    _stack: list[tuple[Symbol, int]]

    def __init__(self, spec: Spec) -> None:
        if __debug__:
            if type(self) == Lr:
                assert spec.pureLR
        assert spec.conflicts == 0
        self._spec = spec
        self._action = spec.actions()
        self._goto = spec.goto()
        self._start_sym = spec.start_sym()
        self.reset()
        self.verbose = False

    @property
    def spec(self) -> Spec:
        return self._spec

    @property
    def start(self) -> list[Symbol] | None:
        """A list of parsing results.  For LR parsing, there is only ever one
        result, but for compatibility with the Glr interface, start is a
        list."""
        return self._start

    def reset(self) -> None:
        self._start = None
        self._stack = [(Epsilon(), 0)]

    def token(self, token: Token) -> None:
        """Feed a token to the parser."""
        tokenSpec = self._spec.sym_spec(token)
        self._act(token, tokenSpec)  # type: ignore

    def eoi(self) -> None:
        """Signal end-of-input to the parser."""
        token = EndOfInput()
        self.token(token)

        assert self._stack[-1][0] == token  # <$>.
        if self.verbose:
            self._printStack()
            print("   --> accept")
        self._stack.pop()

        self._start = [self._stack[1][0]]
        assert self._spec.sym_spec(self._start[0]) == self._start_sym

    def _act(self, sym: Token, symSpec: TokenSpec) -> None:
        if self.verbose:
            self._printStack()
            print("INPUT: %r" % symSpec)

        while True:
            top = self._stack[-1]
            if symSpec not in self._action[top[1]]:
                raise UnexpectedToken("Unexpected token: %r" % sym)

            actions = self._action[top[1]][symSpec]
            assert len(actions) == 1
            action = actions[0]

            if self.verbose:
                print("   --> %r" % action)
            if type(action) == ShiftAction:
                self._stack.append((sym, action.nextState))
                break
            else:
                assert type(action) == ReduceAction
                self._reduce(action.production)

            if self.verbose:
                self._printStack()

    def _printStack(self) -> None:
        print("STACK:", end=" ")
        for node in self._stack:
            symSpec = self._spec.sym_spec(node[0])
            print("%r" % symSpec, end=" ")
        print()
        print("      ", end=" ")
        for node in self._stack:
            symSpec = self._spec.sym_spec(node[0])
            print(
                "%r%s"
                % (
                    node[1],
                    (" " * (len("%r" % symSpec) - len("%r" % node[1]))),
                ),
                end=" ",
            )
        print()

    def _reduce(self, production: Production) -> None:
        nRhs = len(production.rhs)
        rhs = []
        for i in range(len(self._stack) - nRhs, len(self._stack)):
            rhs.append(self._stack[i][0])

        r = self._production(production, rhs)

        for i in range(nRhs):
            self._stack.pop()

        top = self._stack[-1]
        self._stack.append((r, self._goto[top[1]][production.lhs]))

    def _production(
        self, production: Production, rhs: list[Symbol]
    ) -> Nonterm:
        sym = production.lhs.nontermType()
        nRhs = len(rhs)
        assert nRhs == len(production.rhs)
        r = production.method(sym, *rhs)

        # Python's method definition syntax makes returning self from %reduce
        # methods cumbersome, so translate None here.
        if r is None:
            r = sym

        return r



import parsing


class P1(parsing.Precedence):
    "%left p1"


class p2(parsing.Precedence):
    """
    %left
        >p1
    """


# Tokens.
class TokenPlus(parsing.Token):
    "%token plus"


class TokenStar(parsing.Token):
    "%token star"


class TokenLparen(parsing.Token):
    "%token lparen"


class TokenRparen(parsing.Token):
    "%token rparen [p2]"


class TokenId(parsing.Token):
    "%token id"


# Non-terminal definitions.
class NontermE(parsing.Nonterm):
    "%start E"

    def reduceA(self, E, plus, T):
        "%reduce E plus T [p1]"
        self.val = "[%s + %s]" % (E.val, T.val)

    def reduceB(self, T):
        "%reduce T"
        self.val = T.val


class T(parsing.Nonterm):
    "%nonterm"

    def reduceA(self, T, star, F):
        "%reduce T star F"
        self.val = "[%s * %s]" % (T.val, F.val)

    def reduceB(self, F):
        "%reduce F"
        self.val = F.val


class NontermF(parsing.Nonterm):
    """
    %nonterm
    F    [p2]
    """

    def reduceA(self, lparen, E, rparen):
        """%reduce
        lparen E rparen
        """
        self.val = "(%s)" % (E.val,)

    def reduceB(self, id):
        "%reduce id [split]"
        self.val = "ID"



"""
This module contains functionality for extracting a grammar from
classes in a module.
"""
from __future__ import annotations

import types
from parsing.interfaces import SpecSource
from parsing.grammar import (
    PrecedenceSpec,
    PrecedenceRef,
    TokenSpec,
    NontermSpec,
)
from parsing.ast import Token, Nonterm, Precedence
from parsing import introspection
from parsing.errors import SpecError


class ModuleSpecSource(SpecSource):
    """
    ModuleSpecSource scans one or several modules for subclasses of relevant
    classes (Precedence, Token, Nonterm) with specific docstrings.
    """

    def __init__(
        self, modules: types.ModuleType | list[types.ModuleType]
    ) -> None:
        if isinstance(modules, types.ModuleType):
            # Wrap single module in a list.
            modules = [modules]
        self.modules = modules
        items = []
        for module in self.modules:
            for k, v in module.__dict__.items():
                if isinstance(v, type) and isinstance(v.__doc__, str):
                    dirtoks = introspection.parse_docstring(v.__doc__)
                    items.append((module, k, v, dirtoks))
        self.named_objs = items
        self._cache_precedences: list[PrecedenceSpec] | None = None
        self._cache_tokens: list[TokenSpec] | None = None
        self._cache_nonterminals: tuple[
            list[NontermSpec], NontermSpec
        ] | None = None

    def get_precedences(self) -> list[PrecedenceSpec]:
        if self._cache_precedences is not None:
            return self._cache_precedences
        result = []
        for module, k, v, dirtoks in self.named_objs:
            if issubclass(v, Precedence) and dirtoks[0] in [
                "%fail",
                "%nonassoc",
                "%left",
                "%right",
                "%split",
            ]:
                name = k
                relationships = {}
                i = 1
                while i < len(dirtoks):
                    tok = dirtoks[i]
                    m = PrecedenceSpec.assoc_tok_re.match(tok)
                    if m:
                        # Precedence relationship.
                        if m.group(2) in relationships:
                            raise SpecError(
                                "Duplicate precedence "
                                "relationship: %s" % v.__doc__
                            )
                        relationships[m.group(2)] = m.group(1)
                    else:
                        m = NontermSpec.token_re.match(tok)
                        if m:
                            if i != 1:
                                raise SpecError(
                                    "Precedence name must come before "
                                    "relationships: %s" % v.__doc__
                                )
                            name = m.group(1)
                        else:
                            raise SpecError(
                                "Invalid precedence specification: %s"
                                % v.__doc__
                            )
                    i += 1

                prec = PrecedenceSpec(name, dirtoks[0][1:], relationships)
                result.append(prec)
        self._cache_precedences = result
        return result

    def get_tokens(self) -> list[TokenSpec]:
        if self._cache_tokens is not None:
            return self._cache_tokens
        result = []
        for module, k, v, dirtoks in self.named_objs:
            if issubclass(v, Token) and dirtoks[0] in ["%token"]:
                name = k
                prec = None
                i = 1
                while i < len(dirtoks):
                    tok = dirtoks[i]
                    m = NontermSpec.precedence_tok_re.match(tok)
                    if m:
                        if i < len(dirtoks) - 1:
                            raise SpecError(
                                "Precedence must come last in token "
                                "specification: %s" % v.__doc__
                            )
                        prec = PrecedenceRef(m.group(1))
                    else:
                        m = NontermSpec.token_re.match(tok)
                        if m:
                            name = m.group(1)
                        else:
                            raise SpecError(
                                "Invalid token specification: %s" % v.__doc__
                            )
                    i += 1
                if prec is None:
                    prec = PrecedenceRef("none")
                token = TokenSpec(v, name, prec)
                result.append(token)
        self._cache_tokens = result
        return result

    def get_nonterminals(self) -> tuple[list[NontermSpec], NontermSpec]:
        if self._cache_nonterminals is not None:
            return self._cache_nonterminals
        result = []
        startSym: NontermSpec | None = None
        for module, k, v, dirtoks in self.named_objs:
            if issubclass(v, Nonterm) and dirtoks[0] in ["%start", "%nonterm"]:
                nonterm, is_start = NontermSpec.from_class(v)
                result.append(nonterm)

                if is_start:
                    # Start symbol.
                    if startSym is not None:
                        raise SpecError(
                            "Only one start non-terminal allowed: %s"
                            % v.__doc__
                        )
                    startSym = nonterm
        assert startSym is not None
        self._cache_nonterminals = (result, startSym)
        return result, startSym



"""
The classes in this module are used to compute the LR(1) automaton
using Pager's (1977) Practical General Method.
"""
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
)
from mypy_extensions import mypyc_attr

import collections
import inspect
import time
import types
import pickle
import sys

from parsing.errors import SpecError
from parsing import interfaces
from parsing import introspection
from parsing import module_spec
from parsing.ast import Symbol
from parsing.grammar import (
    Item,
    PrecedenceSpec,
    PrecedenceRef,
    Production,
    TokenSpec,
    NontermSpec,
    SymbolSpec,
    EndOfInput,
    eoi,
    Epsilon,
    epsilon,
    NontermStart,
    Action,
    ShiftAction,
    ReduceAction,
)

if TYPE_CHECKING:
    from typing_extensions import Literal
    from parsing.interfaces import SpecSource

    SpecCompatibility = Literal[
        "compatible", "incompatible", "itemsets", "repickle"
    ]

    PickleMode = Literal["r", "w", "rw"]

    ConflictResolution = Literal[
        "neither",  # Discard both.
        "old",  # Keep old.
        "both",  # Keep both.
        "new",  # Keep new.
        "err",  # Unresolvable conflict.
    ]

    ActionState = Dict[SymbolSpec, List[Action]]
    GotoState = Dict[SymbolSpec, int]


_firstSetCache: dict[Tuple[SymbolSpec, ...], frozenset[SymbolSpec]] = {}


def computeFirstSet(s: Tuple[SymbolSpec, ...]) -> frozenset[SymbolSpec]:
    firstSet = _firstSetCache.get(s)
    if firstSet is None:
        # Calculate the first set for the string encoded by the s vector.
        result = set()
        mergeEpsilon = True
        for sym in s:
            hasEpsilon = False
            for elm in sym.firstSet:
                if elm == epsilon:
                    hasEpsilon = True
                else:
                    result.add(elm)
            if not hasEpsilon:
                mergeEpsilon = False
                break
        # Merge epsilon if it was in the first set of every symbol.
        if mergeEpsilon:
            result.add(epsilon)

        # Cache the result.
        _firstSetCache[s] = firstSet = frozenset(result)

    return firstSet


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class ItemSet:
    def __init__(self, items: Iterable[Tuple[Item, set[SymbolSpec]]]) -> None:
        self._kernel: Dict[Item, set[SymbolSpec]] = {}
        self._added: Dict[Item, set[SymbolSpec]] = {}
        self._symMap: dict[SymbolSpec, set[Item]] = {}
        self._all: collections.ChainMap[
            Item, set[SymbolSpec]
        ] = collections.ChainMap(self._kernel, self._added)

        for item, lookahead in items:
            assert item.production.lhs.name == "<S>" or item.dotPos != 0
            self._kernel[item] = set(lookahead)
            sym = item.symbol
            if sym is not None:
                if sym in self._symMap:
                    self._symMap[sym].add(item)
                else:
                    self._symMap[sym] = {item}

        self._hash = hash(frozenset(self._kernel))

    def __repr__(self) -> str:
        kernel = ", ".join(
            self._item_repr(i, la) for i, la in self._kernel.items()
        )
        added = ", ".join(
            self._item_repr(i, la) for i, la in self._added.items()
        )
        return "ItemSet(kernel: %s, added: %r)" % (kernel, added)

    def _item_repr(self, item: Item, lookahead: set[SymbolSpec]) -> str:
        strs = []
        strs.append("[%r ::=" % item.production.lhs)
        assert item.dotPos <= len(item.production.rhs)
        i = 0
        while i < item.dotPos:
            strs.append(" %r" % item.production.rhs[i])
            i += 1
        strs.append(" *")
        while i < len(item.production.rhs):
            strs.append(" %r" % item.production.rhs[i])
            i += 1
        strs.append(
            "., %s] [%s]"
            % (
                "/".join(["%r" % sym for sym in lookahead]),
                item.production.prec.name,
            )
        )

        return "".join(strs)

    def __len__(self) -> int:
        return len(self._kernel)

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: Any) -> bool:
        if type(other) == ItemSet:
            return self._kernel.keys() == other._kernel.keys()
        else:
            return NotImplemented

    def __iter__(self) -> Iterator[Item]:
        return iter(self._all)

    def items(self) -> Iterator[Tuple[Item, set[SymbolSpec]]]:
        return iter(self._all.items())

    # Merge an added item.
    def addedAppend(self, item: Item, lookahead: Iterable[SymbolSpec]) -> bool:
        assert item.dotPos == 0
        assert item.production.lhs.name != "<S>"

        curLookahead = self._added.get(item)
        if curLookahead is not None:
            oldLen = len(curLookahead)
            curLookahead.update(lookahead)
            return oldLen != len(curLookahead)
        else:
            self._added[item] = set(lookahead)
            sym = item.symbol
            if sym is not None:
                if sym in self._symMap:
                    self._symMap[sym].add(item)
                else:
                    self._symMap[sym] = {item}
            return True

    # Given a list of items, compute their closure and merge the results into
    # the set of added items.
    def _closeItems(
        self, items: Iterable[Tuple[Item, Iterable[SymbolSpec]]]
    ) -> None:
        # Iterate over the items until no more can be added to the closure.
        worklist = list(items)
        i = 0
        while i < len(worklist):
            item, lookahead = worklist[i]
            sym = item.symbol
            if isinstance(sym, NontermSpec):
                prefix = item.production.rhs[item.dotPos + 1 :]
                for lookaheadSym in lookahead:
                    firstSet = computeFirstSet(prefix + (lookaheadSym,))
                    for prod in sym.productions:
                        tItem = prod.item(0)
                        if self.addedAppend(tItem, firstSet):
                            worklist.append((tItem, firstSet))
            i += 1

    # Calculate and merge the kernel's transitive closure.
    def closure(self) -> None:
        self._closeItems(self._kernel.items())

    # Calculate the kernel of the goto set, given a particular symbol.
    def goto(self, sym: SymbolSpec) -> ItemSet | None:
        items = self._symMap.get(sym)
        if items:
            return ItemSet([
                (i.production.item(i.dotPos + 1), self._all[i]) for i in items
            ])
        else:
            return None

    # Merge the kernel of other into this ItemSet, then update the closure.
    # It is not sufficient to copy other's added items, since other has not
    # computed its closure.
    def merge(self, other: ItemSet) -> bool:
        items = []
        for item, oLookahead in other._kernel.items():
            sLookahead = self._kernel.get(item)
            # The other ItemSet must be weakly compatible with this one,
            # so it must share the kernel.
            assert sLookahead is not None
            tLookahead = oLookahead - sLookahead
            if tLookahead:
                sLookahead.update(tLookahead)
                items.append((item, tLookahead))

        if items:
            self._closeItems(items)
            return True
        else:
            return False

    # Determine if self and other are weakly compatible, as defined by the
    # Pager(1977) algorithm.
    def weakCompat(self, other: ItemSet) -> bool:
        # Check for identical kernel LR(0) items,
        # and pair items, for later use.
        if len(self) != len(other):
            return False
        pairs = []
        for sItem, sLookahead in self._kernel.items():
            if sItem not in other._kernel:
                return False
            oLookahead = other._kernel[sItem]
            pairs.append((sLookahead, oLookahead))

        # Check for lookahead compatibility.
        for i in range(len(pairs) - 1):
            isLookahead, ioLookahead = pairs[i]
            for j in range(i + 1, len(pairs)):
                jsLookahead, joLookahead = pairs[j]
                if (
                    (
                        not isLookahead.isdisjoint(joLookahead)
                        or not ioLookahead.isdisjoint(jsLookahead)
                    )
                    and isLookahead.isdisjoint(jsLookahead)
                    and ioLookahead.isdisjoint(joLookahead)
                ):
                    return False
        return True


@mypyc_attr(serializable=True, allow_interpreted_subclasses=True)
class Spec(interfaces.Spec):
    """
    The Spec class contains the read-only data structures that the Parser
    class needs in order to parse input.  Parser generation results in a
    Spec instance, which can then be shared by multiple Parser instances."""

    __deletable__ = [
        '_itemSets',
        '_itemSetsHash',
        '_nonterms',
        '_precedences',
        '_productions',
        '_startSym',
        '_startProd',
        '_startState',
        '_tokens',
    ]

    _sym2spec: dict[type[Symbol], SymbolSpec]

    def sym_spec(self, sym: Symbol) -> SymbolSpec:
        return self._sym2spec[type(sym)]

    def actions(self) -> list[ActionState]:
        return self._action

    def goto(self) -> list[GotoState]:
        return self._goto

    def start_sym(self) -> NontermSpec:
        return self._userStartSym

    @property
    def pureLR(self) -> int:
        return self._nConflicts + self._nImpure == 0

    @property
    def conflicts(self) -> int:
        return self._nConflicts

    def __init__(
        self,
        source: types.ModuleType | List[types.ModuleType] | SpecSource,
        pickleFile: Optional[str] = None,
        pickleMode: PickleMode = "rw",
        skinny: bool = True,
        logFile: Optional[str] = None,
        graphFile: Optional[str] = None,
        verbose: bool = False,
        unpickleHook: Callable[[Spec, SpecCompatibility], None] | None = None,
    ) -> None:
        """
        modules : Either a single module, or a list of modules, wherein to
                  look for parser generator directives in docstrings.

        pickleFile : The path of a file to use for Spec pickling/unpickling.

        pickleMode :  "r" : Unpickle from pickleFile.
                      "w" : Pickle to pickleFile.
                      "rw" : Unpickle/pickle from/to pickleFile.

        skinny : If true, discard all data that are only strictly necessary
                 while constructing the parsing tables.  This reduces
                 available debugging context, but substantially reduces
                 pickle size.

        logFile : The path of a file to store a human-readable copy of the
                  parsing tables in.

        graphFile : The path of a file to store a graphviz representation
                    (dot format) of the precedence relationship graph.

        verbose : If true, print progress information while generating the
                  parsing tables."""
        self._skinny = skinny
        self._verbose = verbose

        # Default (no) precedence.
        self._none = PrecedenceSpec("none", "fail", {})
        self._split = PrecedenceSpec("split", "split", {})

        # Symbols are maintained as two separate sets so that non-terminals and
        # terminals (tokens) can be operated on separately where needed.
        self._precedences = {
            self._none.name: self._none,
            self._split.name: self._split,
        }
        self._nonterms: Dict[str, NontermSpec] = {}
        self._tokens: Dict[str, TokenSpec] = {
            eoi.name: eoi,
            epsilon.name: epsilon,
        }
        self._sym2spec: Dict[Type[Symbol], SymbolSpec] = {
            EndOfInput: eoi,
            Epsilon: epsilon,
        }
        self._productions: list[Production] = []

        sources: list[types.ModuleType] | SpecSource
        if isinstance(source, types.ModuleType):
            sources = [source]
        else:
            sources = source

        if not isinstance(sources, list):
            spec_source = sources
        else:
            spec_source = module_spec.ModuleSpecSource(sources)

        if self._verbose:
            print(
                (
                    "Parsing.Spec: Introspecting to acquire formal "
                    "grammar specification..."
                )
            )

        # ===========================================================
        # Precedence.
        #
        for prec in spec_source.get_precedences():
            name = prec.name
            if name in self._precedences:
                raise SpecError("Duplicate precedence name: %s" % (name,))
            if name in self._tokens:
                raise SpecError(
                    "Identical token/precedence names: %s" % (name,)
                )
            if name in self._nonterms:
                raise SpecError(
                    "Identical nonterm/precedence names: %s" % (name,)
                )
            self._precedences[name] = prec

        # ===========================================================
        # Token.
        #
        for token in spec_source.get_tokens():
            name = token.name
            tt = token.tokenType
            if name in self._precedences:
                raise SpecError(
                    "Identical precedence/token names: %s" % tt.__doc__
                )
            if name in self._tokens:
                raise SpecError("Duplicate token name: %s" % tt.__doc__)
            if name in self._nonterms:
                raise SpecError(
                    "Identical nonterm/token names: %s" % tt.__doc__
                )
            self._tokens[name] = token
            self._sym2spec[tt] = token

        # ===========================================================
        # Nonterm.
        #
        nonterms, userStart = spec_source.get_nonterminals()
        for nonterm in nonterms:
            name = nonterm.name
            nt = nonterm.nontermType
            if name in self._precedences:
                raise SpecError(
                    "Identical precedence/nonterm names: %s" % nt.__doc__
                )
            if name in self._tokens:
                raise SpecError(
                    "Identical token/nonterm names: %s" % nt.__doc__
                )
            if name in self._nonterms:
                raise SpecError("Duplicate nonterm name: %s" % nt.__doc__)
            self._nonterms[name] = nonterm
            self._sym2spec[nt] = nonterm

        self._userStartSym = userStart
        if not isinstance(self._userStartSym, NontermSpec):
            raise SpecError("No start symbol specified")

        self._startSym = NontermSpec(
            NontermStart, "<S>", f"{__name__}.NontermStart", self._none
        )

        self._startProd = Production(
            NontermStart.reduce,
            f"{__name__}.NontermStart.reduce",
            self._none,
            self._startSym,
            (self._userStartSym, eoi),
        )

        # Augment grammar with a special start symbol and production:
        #
        #   <S> ::= S <$>.
        self._startSym.productions.add(self._startProd)
        self._nonterms["<S>"] = self._startSym
        self._productions.append(self._startProd)

        # Resolve references in the grammar specification.
        self._references(logFile, graphFile)

        # Everything below this point is computed from the above (once
        # introspection is complete).

        # Each element corresponds to an element in _action.
        self._itemSets: list[ItemSet] = []
        self._itemSetsHash: dict[ItemSet, list[int]] | None = None
        # LR parsing tables.  The tables conceptually contain one state per
        # row, where each row contains one element per symbol.  The table is
        # conceptually in row-major form, but each row is actually a
        # dictionary. If no entry for a symbol exists for a particular state,
        # then input of that symbol is an error for that state.
        self._action: list[ActionState] = []
        self._goto: list[GotoState] = []
        self._startState: int | None = None
        self._nActions = 0
        self._nConflicts = 0
        self._nImpure = 0  # Number of LR impurities (does not affect GLR).

        # Generate parse tables.
        self._prepare(pickleFile, pickleMode, logFile, unpickleHook)

    def __repr__(self) -> str:
        if self._skinny:
            # Print a very reduced summary, since most info has been discarded.
            return "Parsing.Spec: %d states, %d actions (%d split)" % (
                len(self._action),
                self._nActions,
                self._nImpure,
            )

        lines = []

        #  =================================================================
        lines.append("Precedences:")
        deco = [(prec.name, prec) for prec in self._precedences.values()]
        deco.sort()
        for elm in deco:
            prec = elm[1]
            lines.append("  %r" % prec)

        lines.append("Tokens:")
        sym: SymbolSpec
        for token in self._tokens.values():
            lines.append("  %r %r" % (token, token.prec))
            lines.append("    First set: %r" % token.firstSet)
            lines.append("    Follow set: %r" % token.followSet)

        lines.append("Non-terminals:")
        for sym in self._nonterms.values():
            lines.append("  %r %r" % (sym, sym.prec))
            lines.append("    First set: %r" % sym.firstSet)
            lines.append("    Follow set: %r" % sym.followSet)
            lines.append("    Productions:")
            for prod in sym.productions:
                lines.append("      %r" % prod)

        lines.append("Item sets:")
        for i in range(len(self._itemSets)):
            lines.append("  %d: %r" % (i, self._itemSets[i]))
        #  =================================================================

        ntokens = len(self._tokens) - 1
        nnonterms = len(self._nonterms) - 1
        nproductions = len(self._productions) - 1
        nstates = len(self._action)
        lines.append(
            "Parsing.Spec: %d token%s, %d non-terminal%s, "
            "%d production%s, %d state%s, %d action%s (%d split):"
            % (
                ntokens,
                ("s", "")[ntokens == 1],
                nnonterms,
                ("s", "")[nnonterms == 1],
                nproductions,
                ("s", "")[nproductions == 1],
                nstates,
                ("s", "")[nstates == 1],
                self._nActions,
                ("s", "")[self._nActions == 1],
                self._nImpure,
            )
        )
        if self.pureLR:
            lines.append("Algorithm compatibility: GLR, LR")
        elif self._nConflicts == 0:
            lines.append("Algorithm compatibility: GLR")
        else:
            lines.append("Algorithm compatibility: None, due to ambiguity")
        lines.append("Parsing tables:")
        for i in range(len(self._action)):
            lines.append("  %s" % ("=" * 78))
            lines.append(
                "  State %d:%s"
                % (i, ("", " (start state)")[self._startState == i])
            )
            for item in self._itemSets[i]:
                lines.append(
                    " %s%s" % (" " * (len("%d" % i) + 9), item.lr0__repr__())
                )
            lines.append("    Goto:")
            for sym in self._goto[i]:
                lines.append("    %15r : %r" % (sym, self._goto[i][sym]))
            lines.append("    Action:")
            for sym in self._action[i]:
                for action in self._action[i][sym]:
                    conflict = "   "
                    for other in self._action[i][sym]:
                        if action != other:
                            resolution = self._resolve(sym, other, action)
                            if resolution == "err":
                                conflict = "XXX"
                                break

                    if type(action) == ShiftAction:
                        lines.append(
                            "%s %15r : %-6s %d [%s]"
                            % (
                                conflict,
                                sym,
                                "shift",
                                action.nextState,
                                sym.prec.name,
                            )
                        )
                    else:
                        assert type(action) == ReduceAction
                        lines.append(
                            "%s %15r : %-6s %r"
                            % (conflict, sym, "reduce", action.production)
                        )

        ret = "\n".join(lines)
        return ret

    def _prepare(
        self,
        pickleFile: Optional[str],
        pickleMode: PickleMode,
        logFile: Optional[str],
        unpickleHook: Callable[[Spec, SpecCompatibility], None] | None = None,
    ) -> None:
        """
        Compile the specification into data structures that can be used by
        the Parser class for parsing."""
        # Check for a compatible pickle.
        compat = self._unpickle(pickleFile, pickleMode)
        if unpickleHook is not None:
            unpickleHook(self, compat)

        if self._verbose and compat in ["itemsets", "incompatible"]:
            start = time.monotonic()

        if compat == "incompatible":
            # Create the collection of sets of LR(1) items.
            self._firstSets()
            self._followSets()
            self._items()

        if compat == "compatible":
            # Just because the pickle was compatible does not mean that it is
            # valid for parsing.
            if self._nConflicts != 0:
                raise SpecError(
                    "Compatible pickle is invalid due to conflicts (%d)"
                    % self._nConflicts
                )
        if compat in ["itemsets", "incompatible"]:
            # Generate LR(1) parsing tables.
            self._lr()

            # Disambiguate actions.
            self._disambiguate()

            # Check for unused or ambiguous definitions, as well as reporting
            # ambiguities.
            try:
                self._validate(logFile)
            finally:
                if self._verbose:
                    print(
                        "Parsing.Spec: LR(1) parser generation took "
                        f"{(time.monotonic() - start) * 1000:.1f} milliseconds"
                    )
                sys.stdout.flush()
                # Pickle the spec, if method parameters so dictate, even if
                # there were validation errors, so that the pickle might be
                # used in part during later runs.
                self._pickle(pickleFile, pickleMode)
        elif compat == "repickle":
            # Pickle the spec, if method parameters so dictate.
            self._pickle(pickleFile, pickleMode)

        if self._skinny:
            # Discard data that are not needed during parsing.  Note that
            # _pickle() also discarded data that don't even need to be pickled.
            del self._precedences
            del self._nonterms
            del self._tokens
            del self._productions

    # Introspect modules and find special parser declarations.  In order to be
    # a special class, the class must both 1) be subclassed from Token or
    # Nonterm, and 2) contain the appropriate %foo docstring.
    def _introspect(self, spec_source: SpecSource) -> None:
        if self._verbose:
            print(
                (
                    "Parsing.Spec: Introspecting to acquire formal "
                    "grammar specification..."
                )
            )

        self._precedences["none"] = self._none
        self._precedences["split"] = self._split

        # ===========================================================
        # Precedence.
        #
        for prec in spec_source.get_precedences():
            name = prec.name
            if name in self._precedences:
                raise SpecError("Duplicate precedence name: %s" % (name,))
            if name in self._tokens:
                raise SpecError(
                    "Identical token/precedence names: %s" % (name,)
                )
            if name in self._nonterms:
                raise SpecError(
                    "Identical nonterm/precedence names: %s" % (name,)
                )
            self._precedences[name] = prec

        # ===========================================================
        # Token.
        #
        for token in spec_source.get_tokens():
            name = token.name
            tt = token.tokenType
            if name in self._precedences:
                raise SpecError(
                    "Identical precedence/token names: %s" % tt.__doc__
                )
            if name in self._tokens:
                raise SpecError("Duplicate token name: %s" % tt.__doc__)
            if name in self._nonterms:
                raise SpecError(
                    "Identical nonterm/token names: %s" % tt.__doc__
                )
            self._tokens[name] = token
            self._sym2spec[tt] = token

        # ===========================================================
        # Nonterm.
        #
        nonterms, userStart = spec_source.get_nonterminals()
        for nonterm in nonterms:
            name = nonterm.name
            nt = nonterm.nontermType
            if name in self._precedences:
                raise SpecError(
                    "Identical precedence/nonterm names: %s" % nt.__doc__
                )
            if name in self._tokens:
                raise SpecError(
                    "Identical token/nonterm names: %s" % nt.__doc__
                )
            if name in self._nonterms:
                raise SpecError("Duplicate nonterm name: %s" % nt.__doc__)
            self._nonterms[name] = nonterm
            self._sym2spec[nt] = nonterm

        self._userStartSym = userStart
        if not isinstance(self._userStartSym, NontermSpec):
            raise SpecError("No start symbol specified")

    # Resolve all symbolic (named) references.
    def _references(
        self, logFile: Optional[str], graphFile: Optional[str]
    ) -> None:
        # Build the graph of Precedence relationships.
        self._resolvePrec(graphFile)

        # Resolve Token-->Precedence references.
        for token in self._tokens.values():
            if isinstance(token.prec, PrecedenceRef):
                token.prec = self._precedences[token.prec.name]

        # Resolve Nonterm-->Precedence references.
        for nonterm in self._nonterms.values():
            if isinstance(nonterm.prec, PrecedenceRef):
                nonterm.prec = self._precedences[nonterm.prec.name]

        # Resolve Nonterm-->{Nonterm,Token,Precedence} references.
        for nonterm in self._nonterms.values():
            methods = inspect.getmembers(
                nonterm.nontermType,
                predicate=lambda f: (
                    inspect.isfunction(f) and isinstance(f.__doc__, str)
                ),
            )
            for k, v in methods:
                dirtoks = introspection.parse_docstring(v.__doc__)
                if dirtoks[0] == "%reduce":
                    rhs: List[SymbolSpec] = []
                    rhs_terms = []
                    prec = None
                    for i in range(1, len(dirtoks)):
                        tok = dirtoks[i]
                        m = NontermSpec.token_re.match(tok)
                        if m:
                            # Symbolic reference.
                            if tok in self._tokens:
                                rhs.append(self._tokens[tok])
                                rhs_terms.append(self._tokens[tok])
                            elif tok in self._nonterms:
                                rhs.append(self._nonterms[tok])
                            else:
                                raise SpecError(
                                    "Unknown symbol '%s' in reduction "
                                    "specification: %s" % (tok, v.__doc__)
                                )
                        else:
                            m = NontermSpec.precedence_tok_re.match(tok)
                            if m:
                                # Precedence.
                                if i < len(dirtoks) - 1:
                                    raise SpecError(
                                        "Precedence must come last in "
                                        "reduction specification: %s"
                                        % v.__doc__
                                    )
                                if m.group(1) not in self._precedences:
                                    raise SpecError(
                                        "Unknown precedence in reduction "
                                        "specification: %s" % v.__doc__
                                    )
                                prec = self._precedences[m.group(1)]

                    if prec is None:
                        # Inherit the non-terminal's precedence.
                        if rhs_terms:
                            # Inherit the precedence of the last terminal
                            # symbol in rhs
                            prec = rhs_terms[-1].prec
                        else:
                            # Inherit the non-terminal's precedence.
                            prec = nonterm.prec

                    prod = Production(
                        v,
                        "%s.%s" % (nonterm.qualified, k),
                        prec,
                        nonterm,
                        tuple(rhs),
                    )
                    assert prod not in nonterm.productions
                    nonterm.productions.add(prod)
                    self._productions.append(prod)

        if self._verbose:
            ntokens = len(self._tokens) - 1
            nnonterms = len(self._nonterms) - 1
            nproductions = len(self._productions) - 1
            print(
                "Parsing.Spec: %d token%s, "
                "%d non-terminal%s, %d production%s"
                % (
                    ntokens,
                    ("s", "")[ntokens == 1],
                    nnonterms,
                    ("s", "")[nnonterms == 1],
                    nproductions,
                    ("s", "")[nproductions == 1],
                )
            )

    # Build the graph of Precedence relationships.
    def _resolvePrec(self, graphFile: Optional[str]) -> None:
        # Resolve symbolic references and populate equiv/dominators.
        for precA in self._precedences.values():
            for precBName in precA.relationships:
                if precBName not in self._precedences:
                    raise SpecError(
                        (
                            "Precedence '%s' specifies a relationship with "
                            "unknown Precedence '%s'"
                        )
                        % (precA, precBName)
                    )
                precB = self._precedences[precBName]
                rel = precA.relationships[precBName]
                if rel == "=":
                    if precB not in precA.equiv:
                        precA.equiv.add(precB)
                        precA.equiv.update(precB.equiv)
                        precA.dominators.update(precB.dominators)
                        for precC in precB.equiv.copy():
                            precC.equiv = precA.equiv
                            precC.dominators = precA.dominators
                        assert precB.equiv is precA.equiv
                        assert precB.dominators is precA.dominators
                elif rel == "<":
                    precA.dominators.add(precB)
                elif rel == ">":
                    precB.dominators.add(precA)
                else:
                    assert False

        equiv_classes = {}
        for prec in self._precedences.values():
            equiv_classes[id(prec.equiv)] = next(iter(prec.equiv))

        # Write graphviz precedence graph to graphFile, if graphFile was
        # specified.
        if graphFile is not None:
            with open(graphFile, "w+") as f:
                if self._verbose:
                    print(
                        "Parsing.Spec: Writing graphviz "
                        "precedence graph to '%s'..." % graphFile
                    )
                print("digraph Precedence {", file=f)
                print('    graph [bgcolor=black, labeljust="l"]', file=f)
                print(
                    "    node [shape=record, style=filled, color=black, "
                    "fillcolor=gray, fontname=Helvetica, fontsize=10.0]",
                    file=f,
                )
                print("    edge [color=gray]", file=f)
                for precA in self._precedences.values():
                    if precA is next(iter(precA.equiv)):
                        print(
                            '    Precedence_%s [label="{%s}"]'
                            % (
                                precA.name,
                                "\\n".join(
                                    [
                                        "%s (%s)" % (p.name, p.assoc)
                                        for p in precA.equiv
                                    ]
                                ),
                            ),
                            file=f,
                        )
                        for precB in precA.dominators:
                            print(
                                "    Precedence_%s -> Precedence_%s"
                                % (
                                    next(iter(precB.equiv)).name,
                                    next(iter(precA.equiv)).name,
                                ),
                                file=f,
                            )
                print("}", file=f)

        # Iteratively build dominator sets until no more work can be done.
        done = False
        while not done:
            done = True
            for precA in equiv_classes.values():
                for precB in precA.dominators.copy():
                    diff = precB.equiv - precA.dominators
                    if diff:
                        precA.dominators.update(diff)
                        done = False
                    for precC in precB.dominators:
                        diff = precC.equiv - precA.dominators
                        if diff:
                            precA.dominators.update(diff)
                            done = False

        # Check for cycles in the graph.
        cycles = []
        for precA in self._precedences.values():
            for precB in set((precA,)) | precA.equiv:
                if precB in precA.dominators:
                    cycles.append(
                        "Precedence relationship cycle involving '%s'"
                        % precA.name
                    )
        if len(cycles) > 0:
            raise SpecError("\n".join(cycles))

    # Store state to a pickle file, if requested.
    def _pickle(self, file: Optional[str], mode: PickleMode) -> None:
        if self._skinny:
            # Discard data that don't need to be pickled.
            del self._startSym
            del self._startProd
            del self._itemSets
            del self._itemSetsHash
            del self._startState

        if file is not None and "w" in mode:
            if self._verbose:
                print(
                    "Parsing.Spec: Creating %s Spec pickle in %s..."
                    % (("fat", "skinny")[self._skinny], file)
                )

            with open(file, "wb") as f:
                pickle.dump(self, f, protocol=pickle.HIGHEST_PROTOCOL)

    # Restore state from a pickle file, if a compatible one is provided.  This
    # method uses the same set of return values as does _compatible().
    def _unpickle(
        self, file: Optional[str], mode: PickleMode
    ) -> SpecCompatibility:
        if file is not None and "r" in mode:
            if self._verbose:
                print(
                    "Parsing.Spec: Attempting to use pickle from "
                    'file "%s"...' % file
                )
            try:
                with open(file, "rb") as f:
                    # Any exception at all in unpickling can be assumed to be
                    # due to an incompatible pickle.
                    try:
                        spec: Spec = pickle.load(f)
                    except Exception:
                        if self._verbose:
                            error = sys.exc_info()
                            print(
                                "Parsing.Spec: Pickle load failed: "
                                "Exception %s: %s" % (error[0], error[1])
                            )
                        return "incompatible"
            except IOError:
                if self._verbose:
                    error = sys.exc_info()
                    print(
                        "Parsing.Spec: Pickle open failed: "
                        "Exception %s: %s" % (error[0], error[1])
                    )
                return "incompatible"

            compat = self._compatible(spec)
            if compat == "incompatible":
                if self._verbose:
                    print(
                        'Parsing.Spec: Pickle in "%s" '
                        "is incompatible." % file
                    )
                return compat

            if self._verbose:
                print(
                    'Parsing.Spec: Using %s pickle in "%s" (%s)...'
                    % (("fat", "skinny")[spec._skinny], file, compat)
                )

            if compat in ["compatible", "repickle"]:
                # Copy spec's data structures.
                self._precedences = spec._precedences
                self._action = spec._action
                self._goto = spec._goto
                if not self._skinny:
                    self._startState = spec._startState
                self._nActions = spec._nActions
                self._nConflicts = spec._nConflicts
                self._nImpure = spec._nImpure
            elif compat == "itemsets":
                # Precedences are incompatible, so great care has to be taken
                # when copying from the pickle.  Overwrite all precedence
                # specifications in spec with the new ones, then copy over all
                # of the new symbols/productions (but not the new precedences,
                # of course).  This still leaves table generation, which is
                # done by the _prepare() method later.

                # Nonterminals.
                for key in self._nonterms:
                    nontermSelf = self._nonterms[key]
                    nontermSpec = spec._nonterms[key]
                    nontermSpec.prec = nontermSelf.prec
                    # Productions.
                    for prodSelf in nontermSelf.productions:
                        for prodSpec in nontermSpec.productions:
                            if prodSelf.qualified == prodSpec.qualified:
                                prodSpec.prec = prodSelf.prec
                                break
                        assert prodSelf.qualified == prodSpec.qualified
                # Tokens.
                for key in self._tokens:
                    tokenSelf = self._tokens[key]
                    tokenSpec = spec._tokens[key]
                    tokenSpec.prec = tokenSelf.prec
            else:
                assert False

            # Copy spec data structures that are usable regardless of whether
            # the parsing tables need to be rebuilt.
            self._nonterms = spec._nonterms
            self._tokens = spec._tokens
            self._sym2spec = spec._sym2spec
            self._productions = spec._productions
            self._userStartSym = spec._userStartSym
            if not self._skinny:
                self._startSym = spec._startSym
                self._startProd = spec._startProd
                self._itemSets = spec._itemSets
                self._itemSetsHash = spec._itemSetsHash

            return compat
        else:
            return "incompatible"

    # Determine whether other is compatible with self.  Note that self is not
    # completely initialized; the idea here is to determine whether other's
    # data structures can be copied *before* doing the work of building parsing
    # tables.
    #
    # Itemsets and precedences are not directly related, other than that
    # symbols have precedences associated with them.  Therefore, we check for
    # the following cases:
    #
    #   "compatible" : Completely compatible.
    #
    #   "repickle" : Compatible, but pickle needs to be regenerated.
    #
    #   "itemsets" : Itemsets are compatible, but precedence specifications are
    #                not.
    #
    #   "incompatible" : No useful compatibility.
    def _compatible(self, other: Spec) -> SpecCompatibility:
        ret: SpecCompatibility = "compatible"

        if (not self._skinny) and other._skinny:
            return "incompatible"
        elif self._skinny != other._skinny:
            ret = "repickle"

        # Precedences.
        if len(self._precedences) != len(other._precedences):
            if self._verbose:
                print(
                    "Parsing.Spec: Unequal number of precedences (%d vs %d)"
                    % (len(self._precedences), len(other._precedences))
                )
            ret = "itemsets"
        for key in self._precedences:
            if key not in other._precedences:
                if self._verbose:
                    print("Parsing.Spec: Missing precedence: %s" % key)
                ret = "itemsets"
                continue
            precA = self._precedences[key]
            precB = other._precedences[key]
            if (
                precA.name != precB.name
                or precA.assoc != precB.assoc
                or len(precA.relationships) != len(precB.relationships)
            ):
                if self._verbose:
                    print(
                        "Parsing.Spec: Incompatible precedences: %r vs. %r"
                        % (precA, precB)
                    )
                ret = "itemsets"
                continue
            for prec in precA.relationships:
                rel = precA.relationships[prec]
                if (
                    prec not in precB.relationships
                    or precB.relationships[prec] != rel
                ):
                    if self._verbose:
                        print(
                            "Parsing.Spec: Incompatible precedences: "
                            "%r vs. %r" % (precA, precB)
                        )
                    ret = "itemsets"
                    break

        # Nonterminals.
        if len(self._nonterms) != len(other._nonterms):
            if self._verbose:
                print(
                    "Parsing.Spec: Unequal number of non-terminals "
                    "(%d vs %d)" % (len(self._nonterms), len(other._nonterms))
                )
            return "incompatible"
        for key in self._nonterms:
            if key not in other._nonterms:
                if self._verbose:
                    print("Parsing.Spec: Missing non-terminal: %s" % key)
                return "incompatible"
            nontermA = self._nonterms[key]
            nontermB = other._nonterms[key]
            if (
                nontermA.name != nontermB.name
                or nontermA.qualified != nontermB.qualified
                or nontermA.nontermType != nontermB.nontermType
            ):
                if self._verbose:
                    print(
                        "Parsing.Spec: Incompatible non-terminals: "
                        "%r vs. %r" % (nontermA, nontermB)
                    )
                return "incompatible"
            if nontermA.prec.name != nontermB.prec.name:
                if self._verbose:
                    print(
                        (
                            "Parsing.Spec: Differing precedences for "
                            + "non-terminal: %r"
                        )
                        % nontermA
                    )
                ret = "itemsets"

            # Productions.
            if len(nontermA.productions) != len(nontermB.productions):
                if self._verbose:
                    print(
                        "Parsing.Spec: Unequal number of productions "
                        "(%d vs %d)"
                        % (
                            len(self._productions) - 1,
                            len(other._productions) - 1,
                        )
                    )
                return "incompatible"
            for prodA in nontermA.productions:
                match = False
                for prodB in nontermB.productions:
                    if (
                        prodA.qualified == prodB.qualified
                        and prodA.lhs.name == prodB.lhs.name
                        and len(prodA.rhs) == len(prodB.rhs)
                    ):
                        match = True
                        for i in range(len(prodA.rhs)):
                            if prodA.rhs[i].name != prodB.rhs[i].name:
                                match = False
                                if self._verbose:
                                    print(
                                        "Parsing.Spec: Incompatible"
                                        " productions: %r vs. %r"
                                        % (prodA, prodB)
                                    )
                                break
                        if prodA.prec.name != prodB.prec.name:
                            if self._verbose:
                                print(
                                    "Parsing.Spec: Differing precedences "
                                    "for production: %r" % prodA
                                )
                            ret = "itemsets"
                if not match:
                    return "incompatible"

        # Tokens.
        if len(self._tokens) != len(other._tokens):
            if self._verbose:
                print(
                    "Parsing.Spec: Unequal number of tokens (%d vs %d)"
                    % (len(self._tokens), len(other._tokens))
                )
            return "incompatible"
        for key in self._tokens:
            if key not in other._tokens:
                if self._verbose:
                    print("Parsing.Spec: Missing token: %s" % key)
                return "incompatible"
            tokenA = self._tokens[key]
            tokenB = other._tokens[key]
            if (
                tokenA.name != tokenB.name
                or tokenA.tokenType != tokenB.tokenType
            ):
                if self._verbose:
                    print(
                        "Parsing.Spec: Incompatible tokens: %r vs. %r"
                        % (tokenA, tokenB)
                    )
                return "incompatible"
            if tokenA.prec.name != tokenB.prec.name:
                if self._verbose:
                    print(
                        "Parsing.Spec: Differing precedences for token: %r"
                        % tokenA
                    )
                ret = "itemsets"

        # User start symbol.
        if self._userStartSym.name != other._userStartSym.name:
            if self._verbose:
                print(
                    "Parsing.Spec: Differing start symbols: %s vs. %s"
                    % (self._userStartSym.name, other._userStartSym.name)
                )
            return "incompatible"

        if other._skinny and ret == "itemsets":
            # The itemsets have to be regenerated, since they weren't pickled.
            ret = "incompatible"
        return ret

    # Check for unused prececence/token/nonterm/reduce specifications, then
    # throw a SpecError if any ambiguities exist in the grammar.
    def _validate(self, logFile: Optional[str]) -> None:
        if self._verbose:
            print("Parsing.Spec: Validating grammar...")

        lines = []
        if self._nConflicts > 0:
            lines.append(
                "Parsing.Spec: %d unresolvable conflict%s"
                % (self._nConflicts, ("s", "")[self._nConflicts == 1])
            )

        # Previous code guarantees that all precedence/token/nonterm names are
        # unique.  Therefore, we can build a single dictionary here that keys
        # on names.
        used: Dict[str, Any] = {}
        productions = []
        for itemSet in self._itemSets:
            for item, lookahead in itemSet.items():
                productions.append(item.production)
                used[item.production.prec.name] = item.production.prec
                lhs: Tuple[SymbolSpec, ...] = (item.production.lhs,)
                for sym in lhs + item.production.rhs:
                    used[sym.name] = sym
                    used[sym.prec.name] = sym.prec

                for token_spec in lookahead:
                    used[token_spec.prec.name] = token_spec.prec

        nUnused = 0

        # Precedences.
        for prec in self._precedences:
            if prec not in [self._none.name, self._split.name]:
                if prec not in used:
                    nUnused += 1
                    lines.append(
                        "Parsing.Spec: Unused precedence: %r"
                        % self._precedences[prec]
                    )

        # Tokens.
        for token in self._tokens:
            if token not in [eoi.name, epsilon.name]:
                if token not in used:
                    nUnused += 1
                    lines.append(
                        "Parsing.Spec: Unused token: %s" % self._tokens[token]
                    )

        # Nonterms.
        for nonterm in self._nonterms:
            if nonterm not in [self._startSym.name]:
                if nonterm not in used:
                    nUnused += 1
                    lines.append(
                        "Parsing.Spec: Unused nonterm: %s"
                        % self._nonterms[nonterm]
                    )

        # Productions.
        for production in self._productions:
            if production not in productions:
                nUnused += 1
                lines.append(
                    "Parsing.Spec: Unused production: %r" % production
                )

        if nUnused > 0:
            lines.insert(
                (1, 0)[self._nConflicts == 0],
                "Parsing.Spec: %d unused definition%s"
                % (nUnused, ("s", "")[nUnused == 1]),
            )

        # Write to logFile, if one was specified.
        if logFile is not None:
            with open(logFile, "w+") as f:
                if self._verbose:
                    print("Parsing.Spec: Writing log to '%s'..." % logFile)
                f.write("%s" % "\n".join(lines + ["%r" % self]))

        # Conflicts are fatal.
        if self._nConflicts > 0:
            raise SpecError("%s" % ("\n".join(lines)))

        # Make sure to let the user know about unused symbols if verbosity is
        # enabled, and there weren't any conflicts to cause notification via an
        # exception.
        if self._verbose:
            ntokens = len(self._tokens) - 1
            nnonterms = len(self._nonterms) - 1
            nproductions = len(self._productions) - 1
            lines.append(
                "Parsing.Spec: %d token%s, %d non-terminal%s, %d production%s"
                % (
                    ntokens,
                    ("s", "")[ntokens == 1],
                    nnonterms,
                    ("s", "")[nnonterms == 1],
                    nproductions,
                    ("s", "")[nproductions == 1],
                )
            )
            sys.stdout.write("%s\n" % "\n".join(lines))

    # Compute the first sets for all symbols.
    def _firstSets(self) -> None:
        # Terminals.
        # first(X) is X for terminals.
        for sym in self._tokens.values():
            sym.firstSetMerge(sym)

        # Non-terminals.
        #
        # Repeat the following loop until no more symbols can be added to any
        # first set.
        done = False
        while not done:
            done = True
            for name in self._nonterms:
                nonterm = self._nonterms[name]
                for prod in nonterm.productions:
                    mergeEpsilon = True
                    # Iterate through the RHS and merge the first sets into
                    # this symbol's, until a preceding symbol's first set does
                    # not contain epsilon.
                    for elm in prod.rhs:
                        hasEpsilon = False
                        for elmSym in elm.firstSet:
                            if elmSym == epsilon:
                                hasEpsilon = True
                            elif not nonterm.firstSetMerge(elmSym):
                                done = False
                        if not hasEpsilon:
                            mergeEpsilon = False
                            break
                    # Merge epsilon if it was in the first set of every symbol.
                    if mergeEpsilon:
                        if not nonterm.firstSetMerge(epsilon):
                            done = False

    # Compute the follow sets for all symbols.
    def _followSets(self) -> None:
        self._startSym.followSet = {epsilon}

        # Repeat the following loop until no more symbols can be added to any
        # follow set.
        done = False
        while not done:
            done = True
            for name in self._nonterms:
                sym = self._nonterms[name]
                for prod in sym.productions:
                    # For all A ::= aBb, merge first(b) into follow(B).
                    for i in range(len(prod.rhs) - 1):
                        for j in range(i + 1, len(prod.rhs)):
                            if not prod.rhs[i].followSetMerge(
                                prod.rhs[j].firstSet
                            ):
                                done = False
                            if epsilon not in prod.rhs[j].firstSet:
                                break

                    # For A ::= ab, or A ::= aBb where first(b) contains <e>,
                    # merge follow(A) into follow(B).
                    for i in range(len(prod.rhs) - 1, -1, -1):
                        if not prod.rhs[i].followSetMerge(prod.lhs.followSet):
                            done = False
                        if epsilon not in prod.rhs[i].firstSet:
                            break

    # Compute the collection of sets of LR(1) items.
    def _items(self) -> None:
        # Add {[S' ::= * S $., <e>]} to _itemSets.
        tItem = self._startProd.item(0)
        tLookahead: set[SymbolSpec] = {epsilon}
        tItemSet = ItemSet(((tItem, tLookahead),))
        tItemSet.closure()
        self._itemSets.append(tItemSet)
        itemSetCount = 1

        # List of state numbers that need to be processed.
        worklist = [0]
        if self._verbose:
            nwork = len(worklist)
            print(
                "Parsing.Spec: Generating LR(1) itemset collection... ",
                end=" ",
            )
            sys.stdout.write("+")
            sys.stdout.flush()

        # itemSetsHash uses itemsets as keys.  A value is a list of _itemSets
        # indices; these itemsets are the ones referred to the key itemset.
        itemSetsHash = {tItemSet: [0]}

        syms: List[SymbolSpec] = list(self._tokens.values())
        syms += list(self._nonterms.values())
        while worklist:
            if self._verbose:
                if abs(len(worklist) - nwork) >= 10:
                    nwork = len(worklist)
                    sys.stdout.write("[%d/%d]" % (len(worklist), itemSetCount))
                    sys.stdout.flush()

            i = worklist.pop(0)
            itemSet = self._itemSets[i]
            for sym in syms:
                gotoSet = itemSet.goto(sym)
                if gotoSet is None:
                    continue
                merged = False
                if gotoSet in itemSetsHash:
                    for j in itemSetsHash[gotoSet]:
                        mergeSet = self._itemSets[j]
                        if mergeSet.weakCompat(gotoSet):
                            merged = True
                            if mergeSet.merge(gotoSet):
                                # Process worklist in MRU order.  This
                                # causes a depth-first traversal.
                                if j in worklist:
                                    worklist.remove(j)
                                else:
                                    if self._verbose:
                                        sys.stdout.write(".")
                                        sys.stdout.flush()
                                worklist.insert(0, j)
                            break
                if not merged:
                    gotoSet.closure()
                    worklist.append(itemSetCount)
                    if gotoSet not in itemSetsHash:
                        itemSetsHash[gotoSet] = [itemSetCount]
                    else:
                        itemSetsHash[gotoSet].append(itemSetCount)
                    self._itemSets.append(gotoSet)
                    itemSetCount += 1
                    if self._verbose:
                        sys.stdout.write("+")
                        sys.stdout.flush()

        if self._verbose:
            sys.stdout.write("\n")
            sys.stdout.flush()
        self._itemSetsHash = itemSetsHash

    # Compute LR parsing tables.
    def _lr(self) -> None:
        # The collection of sets of LR(1) items already exists.
        assert len(self._itemSets) > 0
        assert len(self._action) == 0
        assert len(self._goto) == 0
        assert self._startState is None
        assert self._nConflicts == 0

        if self._verbose:
            print(
                "Parsing.Spec: Generating LR(1) parsing "
                "tables (%d state%s)... "
                % (len(self._itemSets), ("s", "")[len(self._itemSets) == 1]),
                end=" ",
            )
            sys.stdout.flush()

        itemSetsHash = self._itemSetsHash
        assert itemSetsHash is not None

        for itemSet in self._itemSets:
            if self._verbose:
                sys.stdout.write(".")
                sys.stdout.flush()
            # ==============================================================
            # _action.
            state: ActionState = {}
            self._action.append(state)
            for item, lookahead in itemSet.items():
                # X ::= a*Ab
                sym = item.symbol
                if sym is not None:
                    if isinstance(sym, TokenSpec):
                        itemSetB = itemSet.goto(sym)
                        assert itemSetB is not None
                        for i in itemSetsHash[itemSetB]:
                            itemSetC = self._itemSets[i]
                            if itemSetC.weakCompat(itemSetB):
                                self._actionAppend(state, sym, ShiftAction(i))
                                break

                    # Check if this is the start state.
                    if (
                        self._startState is None
                        and item.production.lhs == self._startSym
                        and item.dotPos == 0
                    ):
                        assert len(item.production.rhs) == 2
                        self._startState = len(self._action) - 1
                # X ::= a*
                else:
                    for lookaheadSym in lookahead:
                        self._actionAppend(
                            state, lookaheadSym, ReduceAction(item.production)
                        )
            # =============================================================
            # _goto.
            gstate: GotoState = {}
            self._goto.append(gstate)
            for nonterm in self._nonterms.values():
                itemSetB = itemSet.goto(nonterm)
                if itemSetB is not None:
                    refs = itemSetsHash.get(itemSetB)
                    if refs:
                        for i in refs:
                            itemSetC = self._itemSets[i]
                            if itemSetC.weakCompat(itemSetB):
                                assert nonterm not in gstate
                                gstate[nonterm] = i
                                break

        if self._verbose:
            sys.stdout.write("\n")
            sys.stdout.flush()

    # Add a symbol action to state, if the action doesn't already exist.
    def _actionAppend(
        self,
        state: dict[SymbolSpec, list[Action]],
        sym: SymbolSpec,
        action: Action,
    ) -> None:
        if sym not in state:
            state[sym] = [action]
        else:
            actions = state[sym]
            if action not in actions:
                state[sym].append(action)

    # Look for action ambiguities and resolve them if possible.
    def _disambiguate(self) -> None:
        assert self._nActions == 0
        assert self._nConflicts == 0
        assert self._nImpure == 0

        if self._verbose:
            print(
                "Parsing.Spec: Disambiguating LR(1) parsing tables... ",
                end=" ",
            )
            sys.stdout.flush()

        for stateInd in range(len(self._action)):
            state = self._action[stateInd]
            if self._verbose:
                vRes = "."
                vNConflicts = 0
            for sym in list(state):
                nConflicts = 0
                acts = [act for act in state[sym]]
                # Construct a list that corresponds to acts; each element
                # indicates whether to preserve the action.
                actStats = [True] * len(acts)

                # Fill in the cells of actStats.
                for i in range(len(acts)):
                    actI = acts[i]
                    for j in range(i + 1, len(acts)):
                        actJ = acts[j]
                        res = self._resolve(sym, actI, actJ)
                        if res == "neither":
                            actStats[i] = False
                            actStats[j] = False
                        elif res == "old":
                            actStats[j] = False
                        elif res == "both":
                            pass
                        elif res == "new":
                            actStats[i] = False
                        elif res == "err":
                            actStats[i] = False
                            actStats[j] = False
                            nConflicts += 1
                        else:
                            assert False

                # Look for actions that can coexist or dominate all other
                # actions.
                newActs = []
                for j in range(len(acts)):
                    if actStats[j]:
                        newActs.append(acts[j])
                # Replace the action set if there exists a valid resolution
                # among the actions.
                if len(newActs) > 0 or nConflicts == 0:
                    if self._verbose:
                        if len(newActs) != len(acts):
                            vRes = "_"
                    state[sym] = newActs
                    nConflicts = 0
                elif self._verbose:
                    vNConflicts += nConflicts

                nActions = len(state[sym])
                if nActions > 1:
                    nImpure = nActions
                else:
                    nImpure = 0

                # Update summary stats.
                self._nActions += nActions
                self._nConflicts += nConflicts
                self._nImpure += nImpure

                # Lr._act() does not expect an empty action list.
                if not state[sym]:
                    state.pop(sym)

            if self._verbose:
                if vNConflicts == 0:
                    sys.stdout.write("%s" % vRes)
                else:
                    sys.stdout.write("[%d:%d]" % (stateInd, vNConflicts))
                sys.stdout.flush()

        if self._verbose:
            sys.stdout.write("\n")
            sys.stdout.flush()

    # Compute how to resolve an action conflict.
    #
    # ret: "neither" : Discard both.
    #      "old"     : Keep old.
    #      "both"    : Keep both.
    #      "new"     : Keep new.
    #      "err"     : Unresolvable conflict.
    def _resolve(
        self, sym: SymbolSpec, oldAct: Action, newAct: Action
    ) -> ConflictResolution:
        ret: ConflictResolution

        if type(oldAct) == ShiftAction:
            oldPrec = sym.prec
        elif type(oldAct) == ReduceAction:
            oldPrec = oldAct.production.prec
        else:
            assert False

        if type(newAct) == ShiftAction:
            newPrec = sym.prec
        elif type(newAct) == ReduceAction:
            newPrec = newAct.production.prec
        else:
            assert False

        if oldPrec in newPrec.dominators:
            # Discard new action.
            ret = "old"
        elif newPrec in oldPrec.dominators:
            # Discard old action.
            ret = "new"
        elif oldPrec in newPrec.equiv:
            assert newPrec in oldPrec.equiv

            if oldPrec.assoc == "split" or newPrec.assoc == "split":
                ret = "both"
            elif type(newAct) == type(oldAct):
                assert type(newAct) == ReduceAction
                assert type(oldAct) == ReduceAction
                # Fatal reduce/reduce conflict.
                ret = "err"
            else:
                if (
                    oldPrec.assoc != "fail"
                    and newPrec.assoc != "fail"
                    and oldPrec.assoc != newPrec.assoc
                ):
                    # Conflicting associativity.
                    ret = "err"
                else:
                    # Determine associativity.  If only one of the actions has
                    # %fail associativity, it is overridden by the other.
                    if oldPrec.assoc == "fail":
                        assoc = newPrec.assoc
                    else:
                        assoc = oldPrec.assoc
                    assert assoc in ["fail", "nonassoc", "left", "right"]

                    if assoc == "fail":
                        ret = "err"
                    elif assoc == "left":
                        if type(oldAct) == ShiftAction:
                            ret = "new"
                        else:
                            assert type(newAct) == ShiftAction
                            ret = "old"
                    elif assoc == "right":
                        if type(oldAct) == ShiftAction:
                            ret = "old"
                        else:
                            assert type(newAct) == ShiftAction
                            ret = "new"
                    elif assoc == "nonassoc":
                        ret = "neither"
                    else:
                        assert False
        else:
            if newPrec in oldPrec.equiv:
                print("%r <--> %r" % (oldPrec, newPrec))
            assert newPrec not in oldPrec.equiv
            # No specified relationship between precedences.
            ret = "err"

        return ret



"""
This module declares several structural ("duck typing") interfaces
that objects or classes can implement to be used in the library
"""

from __future__ import annotations

import abc

from parsing.ast import Symbol, Token
from parsing.grammar import (
    ActionState,
    GotoState,
    SymbolSpec,
    NontermSpec,
    PrecedenceSpec,
    TokenSpec,
)


class Spec(abc.ABC):
    @abc.abstractmethod
    def sym_spec(self, sym: Symbol) -> SymbolSpec:
        raise NotImplementedError

    @abc.abstractmethod
    def actions(self) -> list[ActionState]:
        raise NotImplementedError

    @abc.abstractmethod
    def goto(self) -> list[GotoState]:
        raise NotImplementedError

    @abc.abstractmethod
    def start_sym(self) -> NontermSpec:
        raise NotImplementedError

    @abc.abstractproperty
    def pureLR(self) -> int:
        raise NotImplementedError

    @abc.abstractproperty
    def conflicts(self) -> int:
        raise NotImplementedError


class SpecSource(abc.ABC):
    @abc.abstractmethod
    def get_precedences(self) -> list[PrecedenceSpec]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_tokens(self) -> list[TokenSpec]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_nonterminals(self) -> tuple[list[NontermSpec], NontermSpec]:
        raise NotImplementedError


class Parser(abc.ABC):
    def __init__(self, spec: Spec) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def token(self, token: Token) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def eoi(self) -> None:
        raise NotImplementedError



# ============================================================================
# Copyright (c) 2007 Jason Evans <jasone@canonware.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ============================================================================
#
# Release history:
#
# 1.4 (15 December 2012): Python 3 support.
#
#                         Performance optimizations & bug fixes.
#
#                         Published on PyPI.
#
# 1.3 (8 August 2007): Retroactively number public releases.
#
#                      Back-port to Python 2.4.
#
#                      Remove some magic surrounding epsilon, in order to
#                      generalize/simplify.
#
# 1.2 (6 May 2007): Fix some off-by-one errors in production count reporting.
#
#                   Add some missing code that helps detect which definitions
#                   are used/unused when building the parser.
#
# 1.1 (22 March 2007): Optimize/generalize Lr._production() by using argument
#                      list expansion.
#
# 1.0 (19 March 2007): Initial public release.
#
# ============================================================================
"""
The Parsing module implements an LR(1) parser generator, as well as the
runtime support for using a generated parser, via the Lr and Glr parser
drivers.  There is no special parser generator input file format, but the
parser generator still needs to know what classes/methods correspond to
various aspects of the parser.  This information is specified via
docstrings, which the parser generator introspects in order to generate a
parser.  Only one parser specification can be embedded in each module, but
it is possible to share modules between parser specifications so that, for
example, the same token definitions can be used by multiple parser
specifications.

The parsing tables are LR(1), but they are generated using a fast algorithm
that avoids creating duplicate states that result when using the generic
LR(1) algorithm.  Creation time and table size are on par with the LALR(1)
algorithm.  However, LALR(1) can create reduce/reduce conflicts that don't
exist in a true LR(1) parser.  For more information on the algorithm, see:

    A Practical General Method for Constructing LR(k) Parsers
    David Pager
    Acta Informatica 7, 249-268 (1977)

Parsing table generation requires non-trivial amounts of time for large
grammars.  Internal pickling support makes it possible to cache the most
recent version of the parsing table on disk, and use the table if the
current parser specification is still compatible with the one that was used
to generate the pickled parsing table.  Since the compatibility checking is
quite fast, even for large grammars, this removes the need to use the
standard code generation method that is used by most parser generators.

Parser specifications are encapsulated by the Spec class.  Parser instances
use Spec instances, but are themselves based on separate classes.  This
allows multiple parser instances to exist simultaneously, without requiring
multiple copies of the parsing tables.  There are two separate parser driver
classes:

  Lr : Standard Characteristic Finite State Machine (CFSM) driver, based on
       unambiguous LR(1) parsing tables.  This driver is faster than the Glr
       driver, but it cannot deal with all parsing tables that the Glr
       driver can.

  Glr : Generalized LR driver, capable of tracking multiple parse trees
        simultaneously, if the %split precedence is used to mark ambiguous
        actions.  This driver is closely based on Elkhound's design, which
        is described in a technical report:

            Elkhound: A Fast, Practical GLR Parser Generator
            Scott McPeak
            Report No. UCB/CSD-2-1214 (December 2002)
            http://www.cs.berkeley.edu/~smcpeak/elkhound/

Parser generator directives are embedded in docstrings, and must begin with
a '%' character, followed immediately by one of several keywords:

    Precedence : %fail %nonassoc %left %right %split
         Token : %token
  Non-terminal : %start %nonterm
    Production : %reduce

All of these directives are associated with classes except for %reduce.
%reduce is associated with methods within non-terminal classes.  The Parsing
module provides base classes from which precedences, tokens, and
non-terminals must be derived.  This is not as restrictive as it sounds,
since there is nothing preventing, for example, a master Token class that
subclasses Parsing.Token, which all of the actual token types then subclass.
Also, nothing prevents using multiple inheritance.

Following are the base classes to be subclassed by parser specifications:

  * Precedence
  * Token
  * Nonterm
"""

from __future__ import annotations


__all__ = (
    "Glr",
    "Lr",
    "ModuleSpecSource",
    "Nonterm",
    "Parser",
    "Precedence",
    "Spec",
    "SpecError",
    "SpecSource",
    "UnexpectedToken",
    "Token",
    "__version__",
)

from parsing._version import __version__
from parsing.automaton import Spec
from parsing.ast import Nonterm, Token, Precedence
from parsing.errors import SpecError, UnexpectedToken
from parsing.interfaces import Parser, SpecSource
from parsing.module_spec import ModuleSpecSource
from parsing.lrparser import Lr
from parsing.glrparser import Glr



import os
import pathlib
import sys

from setuptools import extension as setuptools_ext
from setuptools import setup
from setuptools.command import build_ext as setuptools_build_ext


_ROOT = pathlib.Path(__file__).parent


with open(str(_ROOT / "README.rst")) as f:
    readme = f.read()


with open(str(_ROOT / "parsing" / "_version.py")) as f:
    for line in f:
        if line.startswith("__version__ ="):
            _, _, version = line.partition("=")
            VERSION = version.strip(" \n'\"")
            break
    else:
        raise RuntimeError(
            "unable to read the version from parsing/_version.py"
        )


USE_MYPYC = False
MYPY_DEPENDENCY = "mypy>=1.4.1"
setup_requires = []
ext_modules = []

if (
    os.environ.get("PARSING_USE_MYPYC", None) in {"true", "1", "on"}
    or "--use-mypyc" in sys.argv
):
    setup_requires.append(MYPY_DEPENDENCY)
    # Fool setuptools into calling build_ext.  The actual list of
    # extensions would get replaced by mypycify.
    ext_modules.append(
        setuptools_ext.Extension("parsing.foo", ["parsing/foo.c"])
    )
    USE_MYPYC = True


class build_ext(setuptools_build_ext.build_ext):  # type: ignore
    def finalize_options(self) -> None:
        # finalize_options() may be called multiple times on the
        # same command object, so make sure not to override previously
        # set options.
        if getattr(self, "_initialized", False):
            return

        if USE_MYPYC:
            import pkg_resources

            # Double check Cython presence in case setup_requires
            # didn't go into effect (most likely because someone
            # imported Cython before setup_requires injected the
            # correct egg into sys.path.
            try:
                import mypy.version
                from mypyc.build import mypycify
            except ImportError:
                raise RuntimeError(
                    "please install {} to compile parsing from source".format(
                        MYPY_DEPENDENCY
                    )
                )

            mypy_dep = pkg_resources.Requirement.parse(MYPY_DEPENDENCY)
            if mypy.version.__version__ not in mypy_dep:
                raise RuntimeError(
                    "parsing requires {}, got mypy=={}".format(
                        MYPY_DEPENDENCY, mypy.version.__version__
                    )
                )

            self.distribution.ext_modules = mypycify(
                [
                    "parsing/automaton.py",
                    "parsing/grammar.py",
                ],
            )

        super(build_ext, self).finalize_options()


setup(
    name="parsing",
    version=VERSION,
    python_requires=">=3.7.0",
    url="http://www.canonware.com/Parsing/",
    license="MIT",
    author="Jason Evans",
    author_email="jasone@canonware.com",
    description="A pure-Python module that implements an LR(1) "
    "parser generator, as well as CFSM and GLR parser drivers.",
    long_description=readme,
    long_description_content_type="text/x-rst",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Compilers",
        "Topic :: Text Processing :: General",
    ],
    packages=["parsing", "parsing.tests", "parsing.tests.specs"],
    package_data={"parsing": ["py.typed"]},
    setup_requires=setup_requires,
    ext_modules=ext_modules,
    extras_require={
        "test": [
            "flake8",
            MYPY_DEPENDENCY,
        ]
    },
    install_requires=["mypy-extensions>=1.0.0"],
    cmdclass={"build_ext": build_ext},
)
